# -*- coding: utf-8 -*-
"""
QA/QC CLI utility functions.
"""
import logging
import os

import click
import numpy as np
from gaps.cli import CLICommandFromFunction, as_click_command
from rex.utilities.cli_dtypes import INT, STR, STRLIST
from rex.utilities.loggers import init_logger

from reV import __version__
from reV.qa_qc.qa_qc import QaQc, QaQcModule
from reV.qa_qc.summary import (
    ExclusionsMask,
    SummarizeH5,
    SummarizeSupplyCurve,
    SupplyCurvePlot,
)
from reV.utilities import ModuleName, SupplyCurveField

logger = logging.getLogger(__name__)


def cli_qa_qc(modules, out_dir, max_workers=None):
    """Run QA/QC on reV outputs

    ``reV`` QA/QC performs quality assurance checks on ``reV`` output
    data. Users can specify the type of QA/QC that should be applied
    to each ``reV`` module.

    Parameters
    ----------
    modules : dict
        Dictionary of modules to QA/QC. Keys should be the names of the
        modules to QA/QC. The values are dictionaries that represent the
        config for the respective QA/QC step. Allowed config keys for
        QA/QC are the "property" attributes of
        :class:`~reV.qa_qc.qa_qc.QaQcModule`.
    out_dir : str
        Path to output directory.
    max_workers : int, optional
        Max number of workers to run for QA/QA. If ``None``, uses all
        CPU cores. By default, ``None``.

    Raises
    ------
    ValueError
        If fpath is not an H5 or CSV file.
    """
    for module, mcf in modules.items():
        module_config = QaQcModule(module, mcf, out_dir)

        qa_dir = out_dir
        if module_config.sub_dir is not None:
            qa_dir = os.path.join(out_dir, module_config.sub_dir)

        if module.lower() == 'exclusions':
            QaQc.exclusions_mask(module_config.fpath, qa_dir,
                                 layers_dict=module_config.excl_dict,
                                 min_area=module_config.min_area,
                                 kernel=module_config.area_filter_kernel,
                                 plot_type=module_config.plot_type,
                                 cmap=module_config.cmap,
                                 plot_step=module_config.plot_step)

        elif module_config.fpath.endswith('.h5'):
            QaQc.h5(module_config.fpath, qa_dir, dsets=module_config.dsets,
                    group=module_config.group,
                    process_size=module_config.process_size,
                    max_workers=max_workers,
                    plot_type=module_config.plot_type, cmap=module_config.cmap)

        elif module_config.fpath.endswith('.csv'):
            QaQc.supply_curve(module_config.fpath, qa_dir,
                              columns=module_config.columns,
                              lcoe=module_config.lcoe,
                              plot_type=module_config.plot_type,
                              cmap=module_config.cmap)
        else:
            msg = ("Cannot run QA/QC for {}: 'fpath' must be a '*.h5' "
                   "or '*.csv' reV output file, but {} was given!"
                   .format(module, module_config.fpath))
            logger.error(msg)
            raise ValueError(msg)


qa_qc_command = CLICommandFromFunction(cli_qa_qc, name=str(ModuleName.QA_QC),
                                       split_keys=None)
main = as_click_command(qa_qc_command)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def qa_qc_extra(ctx, verbose):
    """Execute extra QA/QC utility"""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@qa_qc_extra.group(chain=True)
@click.option('--out_dir', '-o', type=click.Path(), required=True,
              help="Directory path to save summary tables and plots too")
@click.option('--log_file', '-log', type=click.Path(), default=None,
              show_default=True,
              help='File to log to, by default None')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def summarize(ctx, out_dir, log_file, verbose):
    """
    Summarize reV data
    """
    ctx.obj['OUT_DIR'] = out_dir
    if any([verbose, ctx.obj['VERBOSE']]):
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    init_logger('reV', log_file=log_file, log_level=log_level)


@summarize.command()
@click.option('--h5_file', '-h5', type=click.Path(exists=True), required=True,
              help='Path to .h5 file to summarize')
@click.option('--dsets', '-ds', type=STRLIST, default=None,
              show_default=True,
              help='Datasets to summarize, by default None')
@click.option('--group', '-grp', type=STR, default=None,
              show_default=True,
              help=('Group within h5_file to summarize datasets for, by '
                    'default None'))
@click.option('--process_size', '-ps', type=INT, default=None,
              show_default=True,
              help='Number of sites to process at a time, by default None')
@click.option('--max_workers', '-w', type=INT, default=None,
              show_default=True,
              help=('Number of workers to use when summarizing 2D datasets,'
                    ' by default None'))
@click.pass_context
def h5(ctx, h5_file, dsets, group, process_size, max_workers):
    """
    Summarize datasets in .h5 file
    """
    SummarizeH5.run(h5_file, ctx.obj['OUT_DIR'], group=group, dsets=dsets,
                    process_size=process_size, max_workers=max_workers)


@summarize.command()
@click.option('--plot_type', '-plt', default='plotly',
              type=click.Choice(['plot', 'plotly'], case_sensitive=False),
              show_default=True,
              help=(" plot_type of plot to create 'plot' or 'plotly', by "
                    "default 'plot'"))
@click.option('--cmap', '-cmap', type=str, default='viridis',
              show_default=True,
              help="Colormap name, by default 'viridis'")
@click.pass_context
def scatter_plots(ctx, plot_type, cmap):
    """
    create scatter plots from h5 summary tables
    """
    QaQc.create_scatter_plots(ctx.obj['OUT_DIR'], plot_type, cmap)


@summarize.command()
@click.option('--sc_table', '-sct', type=click.Path(exists=True),
              required=True, help='Path to .csv containing Supply Curve table')
@click.option('--columns', '-cols', type=STRLIST, default=None,
              show_default=True,
              help=('Column(s) to summarize, if None summarize all numeric '
                    'columns, by default None'))
@click.pass_context
def supply_curve_table(ctx, sc_table, columns):
    """
    Summarize Supply Curve Table
    """
    ctx.obj['SC_TABLE'] = sc_table
    SummarizeSupplyCurve.run(sc_table, ctx.obj['OUT_DIR'], columns=columns)


@summarize.command()
@click.option('--sc_table', '-sct', type=click.Path(exists=True), default=None,
              show_default=True,
              help=("Path to .csv containing Supply Curve table, can be "
                    "supplied in 'supply-curve-table'"))
@click.option('--plot_type', '-plt', default='plotly',
              type=click.Choice(['plot', 'plotly'], case_sensitive=False),
              show_default=True,
              help=(" plot_type of plot to create 'plot' or 'plotly', by "
                    "default 'plot'"))
@click.option('--lcoe', '-lcoe', type=STR, default=SupplyCurveField.MEAN_LCOE,
              help="LCOE value to plot, by default %(default)s")
@click.pass_context
def supply_curve_plot(ctx, sc_table, plot_type, lcoe):
    """
    Plot Supply Curve (cumulative capacity vs LCOE)
    """
    if sc_table is None:
        sc_table = ctx.obj['SC_TABLE']

    SupplyCurvePlot.plot(sc_table, ctx.obj['OUT_DIR'],
                         plot_type=plot_type, lcoe=lcoe)


@summarize.command()
@click.option('--excl_mask', '-mask', type=click.Path(exists=True),
              required=True,
              help='Path to .npy file containing final exclusions mask')
@click.option('--plot_type', '-plt', default='plotly',
              type=click.Choice(['plot', 'plotly'], case_sensitive=False),
              show_default=True,
              help=(" plot_type of plot to create 'plot' or 'plotly', by "
                    "default 'plot'"))
@click.option('--cmap', '-cmap', type=str, default='viridis',
              show_default=True,
              help="Colormap name, by default 'viridis'")
@click.option('--plot_step', '-step', type=int, default=100,
              show_default=True,
              help="Step between points to plot")
@click.pass_context
def exclusions_mask(ctx, excl_mask, plot_type, cmap, plot_step):
    """
    create heat map of exclusions mask
    """
    excl_mask = np.load(excl_mask)
    ExclusionsMask.plot(excl_mask, ctx.obj['OUT_DIR'],
                        plot_type=plot_type, cmap=cmap,
                        plot_step=plot_step)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV QA/QC CLI.')
        raise



# -*- coding: utf-8 -*-
"""
reV Supply Curve CLI utility functions.
"""
import logging
from warnings import warn

from gaps.cli import as_click_command, CLICommandFromClass
from gaps.pipeline import parse_previous_status

from reV.supply_curve.supply_curve import SupplyCurve
from reV.utilities.exceptions import PipelineError
from reV.utilities import ModuleName


logger = logging.getLogger(__name__)


def _preprocessor(config, out_dir):
    """Preprocess supply curve config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    out_dir : str
        Path to output file directory.

    Returns
    -------
    dict
        Updated config file.
    """
    if config.get("sc_points") == 'PIPELINE':
        sc_points = parse_previous_status(out_dir, ModuleName.SUPPLY_CURVE)
        if not sc_points:
            raise PipelineError('Could not parse "sc_points" from previous '
                                'pipeline jobs.')
        config["sc_points"] = sc_points[0]
        logger.info('Supply curve using the following '
                    'pipeline input for sc_points: {}'
                    .format(config["sc_points"]))

    if config.get("simple"):
        no_effect = [key for key in ['avail_cap_frac', 'line_limited']
                     if key in config]
        if no_effect:
            msg = ('The following key(s) have no effect when running '
                   'supply curve with "simple=True": "{}". To silence this '
                   'warning, please remove them from the config'
                   .format(', '.join(no_effect)))
            logger.warning(msg)
            warn(msg)

    return config


sc_command = CLICommandFromClass(SupplyCurve, method="run",
                                 name=str(ModuleName.SUPPLY_CURVE),
                                 add_collect=False, split_keys=None,
                                 config_preprocessor=_preprocessor)
main = as_click_command(sc_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Supply Curve CLI.')
        raise



# -*- coding: utf-8 -*-
"""
Classes to handle reV h5 output files.
"""
import json
import logging
import sys

import NRWAL
import PySAM
import rex
from rex.outputs import Outputs as rexOutputs

from reV.version import __version__

logger = logging.getLogger(__name__)


class Outputs(rexOutputs):
    """
    Base class to handle reV output data in .h5 format

    Examples
    --------
    The reV Outputs handler can be used to initialize h5 files in the standard
    reV/rex resource data format.

    >>> from reV import Outputs
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> meta = pd.DataFrame({SupplyCurveField.LATITUDE: np.ones(100),
    >>>                      SupplyCurveField.LONGITUDE: np.ones(100)})
    >>>
    >>> time_index = pd.date_range('20210101', '20220101', freq='1h',
    >>>                            closed='right')
    >>>
    >>> with Outputs('test.h5', 'w') as f:
    >>>     f.meta = meta
    >>>     f.time_index = time_index

    You can also use the Outputs handler to read output h5 files from disk.
    The Outputs handler will automatically parse the meta data and time index
    into the expected pandas objects (DataFrame and DatetimeIndex,
    respectively).

    >>> with Outputs('test.h5') as f:
    >>>     print(f.meta.head())
    >>>
         latitude  longitude
    gid
    0         1.0        1.0
    1         1.0        1.0
    2         1.0        1.0
    3         1.0        1.0
    4         1.0        1.0

    >>> with Outputs('test.h5') as f:
    >>>     print(f.time_index)
    DatetimeIndex(['2021-01-01 01:00:00+00:00', '2021-01-01 02:00:00+00:00',
                   '2021-01-01 03:00:00+00:00', '2021-01-01 04:00:00+00:00',
                   '2021-01-01 05:00:00+00:00', '2021-01-01 06:00:00+00:00',
                   '2021-01-01 07:00:00+00:00', '2021-01-01 08:00:00+00:00',
                   '2021-01-01 09:00:00+00:00', '2021-01-01 10:00:00+00:00',
                   ...
                   '2021-12-31 15:00:00+00:00', '2021-12-31 16:00:00+00:00',
                   '2021-12-31 17:00:00+00:00', '2021-12-31 18:00:00+00:00',
                   '2021-12-31 19:00:00+00:00', '2021-12-31 20:00:00+00:00',
                   '2021-12-31 21:00:00+00:00', '2021-12-31 22:00:00+00:00',
                   '2021-12-31 23:00:00+00:00', '2022-01-01 00:00:00+00:00'],
                  dtype='datetime64[ns, UTC]', length=8760, freq=None)

    There are a few ways to use the Outputs handler to write data to a file.
    Here is one example using the pre-initialized file we created earlier.
    Note that the Outputs handler will automatically scale float data using
    the "scale_factor" attribute. The Outputs handler will unscale the data
    while being read unless the unscale kwarg is explicityly set to False.
    This behavior is intended to reduce disk storage requirements for big
    data and can be disabled by setting dtype=np.float32 or dtype=np.float64
    when writing data.

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='dset1',
    >>>                     dset_data=np.ones((8760, 100)) * 42.42,
    >>>                     attrs={'scale_factor': 100}, dtype=np.int32)


    >>> with Outputs('test.h5') as f:
    >>>     print(f['dset1'])
    >>>     print(f['dset1'].dtype)
    [[42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     ...
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]]
    float32

    >>> with Outputs('test.h5', unscale=False) as f:
    >>>     print(f['dset1'])
    >>>     print(f['dset1'].dtype)
    [[4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     ...
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]]
    int32

    Note that the reV Outputs handler is specifically designed to read and
    write spatiotemporal data. It is therefore important to intialize the meta
    data and time index objects even if your data is only spatial or only
    temporal. Furthermore, the Outputs handler will always assume that 1D
    datasets represent scalar data (non-timeseries) that corresponds to the
    meta data shape, and that 2D datasets represent spatiotemporal data whose
    shape corresponds to (len(time_index), len(meta)). You can see these
    constraints here:

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='bad_shape',
                            dset_data=np.ones((1, 100)) * 42.42,
                            attrs={'scale_factor': 100}, dtype=np.int32)
    HandlerValueError: 2D data with shape (1, 100) is not of the proper
    spatiotemporal shape: (8760, 100)

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='bad_shape',
                            dset_data=np.ones((8760,)) * 42.42,
                            attrs={'scale_factor': 100}, dtype=np.int32)
    HandlerValueError: 1D data with shape (8760,) is not of the proper
    spatial shape: (100,)
    """

    @property
    def full_version_record(self):
        """Get record of versions for dependencies

        Returns
        -------
        dict
            Dictionary of package versions for dependencies
        """
        rev_versions = {'reV': __version__,
                        'rex': rex.__version__,
                        'pysam': PySAM.__version__,
                        'python': sys.version,
                        'nrwal': NRWAL.__version__,
                        }
        versions = super().full_version_record
        versions.update(rev_versions)
        return versions

    def set_version_attr(self):
        """Set the version attribute to the h5 file."""
        self.h5.attrs['version'] = __version__
        self.h5.attrs['full_version_record'] = json.dumps(
            self.full_version_record)
        self.h5.attrs['package'] = 'reV'



# -*- coding: utf-8 -*-
"""
a simple genetic algorithm
"""
import numpy as np
import time
from math import log
import logging

logger = logging.getLogger(__name__)


class GeneticAlgorithm():
    """a simple genetic algorithm used to select bespoke turbine locations
    """

    def __init__(self, bits, bounds, variable_type, objective_function,
                 max_generation=100, population_size=0, crossover_rate=0.1,
                 mutation_rate=0.01, tol=1E-6, convergence_iters=5,
                 max_time=3600):
        """
        Parameters
        ----------
        bits : array of ints
            The number of bits assigned to each of the design variables.
            The number of discretizations for each design variables will be
            2^n where n is the number of bits assigned to that variable.
        bounds : array of tuples
            The bounds for each design variable. This parameter looks like:
            np.array([(lower, upper), (lower, upper)...])
        variable_type : array of strings ('int' or 'float')
            The type of each design variable (int or float).
        objective_function : function handle for the objective that is to be
            minimized. Should take a single variable as an input which is a
            list/array of the design variables.
        max_generation : int, optional
            The maximum number of generations that will be run in the genetic
            algorithm.
        population_size : int, optional
            The population size in the genetic algorithm.
        crossover_rate : float, optional
            The probability of crossover for a single bit during the crossover
            phase of the genetic algorithm.
        mutation_rate : float, optional
            The probability of a single bit mutating during the mutation phase
            of the genetic algorithm.
        tol : float, optional
            The absolute tolerance to determine convergence.
        convergence_iters : int, optional
            The number of generations to determine convergence.
        max_time : float
            The maximum time (in seconds) to run the genetic algorithm.
        """

        logger.debug('Initializing GeneticAlgorithm...')
        logger.debug('Minimum convergence iterations: {}'
                     .format(convergence_iters))
        logger.debug('Max iterations (generations): {}'.format(max_generation))
        logger.debug('Population size: {}'.format(population_size))
        logger.debug('Crossover rate: {}'.format(crossover_rate))
        logger.debug('Mutation rate: {}'.format(mutation_rate))
        logger.debug('Convergence tolerance: {}'.format(tol))
        logger.debug('Maximum runtime (in seconds): {}'.format(max_time))

        # inputs
        self.bits = bits
        self.bounds = bounds
        self.variable_type = variable_type
        self.objective_function = objective_function
        self.max_generation = max_generation
        self.population_size = population_size
        self.crossover_rate = crossover_rate
        self.mutation_rate = mutation_rate
        self.tol = tol
        self.convergence_iters = convergence_iters
        self.max_time = max_time

        # internal variables, you could output some of this info if you wanted
        self.design_variables = np.array([])  # the desgin variables as they
        # are passed into self.objective function
        self.nbits = 0  # the total number of bits in each chromosome
        self.nvars = 0  # the total number of design variables
        self.parent_population = np.array([])  # 2D array containing all of the
        # parent individuals
        self.offspring_population = np.array([])  # 2D array containing all of
        # the offspring individuals
        self.parent_fitness = np.array([])  # array containing all of the
        # parent fitnesses
        self.offspring_fitness = np.array([])  # array containing all of the
        # offspring fitnesses
        self.discretized_variables = {}  # a dict of arrays containing all of
        # the discretized design variable

        # outputs
        self.solution_history = np.array([])
        self.optimized_function_value = 0.0
        self.optimized_design_variables = np.array([])

        self.initialize_design_variables()
        self.initialize_bits()
        if self.population_size % 2 == 1:
            self.population_size += 1
        self.initialize_population()
        self.initialize_fitness()

        if self.population_size > 5:
            n = 5
        else:
            n = self.population_size
        logger.debug('The first few parent individuals are: {}'
                     .format(self.parent_population[0:n]))
        logger.debug('The first few parent fitness values are: {}'
                     .format(self.parent_fitness[0:n]))

    def initialize_design_variables(self):
        """initialize the design variables from the randomly initialized
        population
        """
        # determine the number of design variables and initialize
        self.nvars = len(self.variable_type)
        self.design_variables = np.zeros(self.nvars)
        float_ind = 0
        for i in range(self.nvars):
            if self.variable_type[i] == "float":
                ndiscretizations = 2**self.bits[i]
                self.discretized_variables["float_var%s" % float_ind] = \
                    np.linspace(self.bounds[i][0], self.bounds[i][1],
                                ndiscretizations)
                float_ind += 1

    def initialize_bits(self):
        """determine the total number of bits"""
        # determine the total number of bits
        for i in range(self.nvars):
            if self.variable_type[i] == "int":
                int_range = self.bounds[i][1] - self.bounds[i][0]
                int_bits = int(np.ceil(log(int_range, 2)))
                self.bits[i] = int_bits
            self.nbits += self.bits[i]

    def initialize_population(self):
        """randomly initialize the parent and offspring populations"""
        all_bits_on = np.ones((1, self.nbits))
        random_bits_on = np.random.randint(
            0, high=2, size=(self.population_size - 1, self.nbits)
        )
        self.parent_population = np.r_[all_bits_on, random_bits_on]
        self.offspring_population = np.zeros_like(self.parent_population)

    def initialize_fitness(self):
        """initialize the fitness of member of the parent population"""
        # initialize the fitness arrays
        self.parent_fitness = np.zeros(self.population_size)
        self.offspring_fitness = np.zeros(self.population_size)

        # initialize fitness of the parent population
        for i in range(self.population_size):
            self.chromosome_2_variables(self.parent_population[i])
            self.parent_fitness[i] = \
                self.objective_function(self.design_variables)

    def chromosome_2_variables(self, chromosome):
        """convert the binary chromosomes to design variable values"""

        first_bit = 0
        float_ind = 0

        for i in range(self.nvars):
            binary_value = 0
            for j in range(self.bits[i]):
                binary_value += chromosome[first_bit + j] * 2**j
            first_bit += self.bits[i]

            if self.variable_type[i] == "float":
                self.design_variables[i] = \
                    self.discretized_variables["float_var%s"
                                               % float_ind][binary_value]
                float_ind += 1

            elif self.variable_type[i] == "int":
                self.design_variables[i] = self.bounds[i][0] + binary_value

    def crossover(self):
        """perform crossover between individual parents"""
        self.offspring_population[:, :] = self.parent_population[:, :]

        # mate conscutive pairs of parents (0, 1), (2, 3), ...
        # The population is shuffled so this does not need to be randomized
        for i in range(int(self.population_size / 2)):
            # trade bits in the offspring
            crossover_arr = np.random.rand(self.nbits)
            for j in range(self.nbits):
                if crossover_arr[j] < self.crossover_rate:
                    self.offspring_population[2 * i][j], \
                        self.offspring_population[2 * i + 1][j] = \
                        self.offspring_population[2 * i + 1][j], \
                        self.offspring_population[2 * i][j]

    def mutate(self):
        """randomly mutate bits of each chromosome"""
        for i in range(int(self.population_size)):
            # mutate bits in the offspring
            mutate_arr = np.random.rand(self.nbits)
            for j in range(self.nbits):
                if mutate_arr[j] < self.mutation_rate:
                    self.offspring_population[i][j] = \
                        (self.offspring_population[i][j] + 1) % 2

    def optimize_ga(self):
        """run the genetic algorithm"""

        converged = False
        ngens = 1
        generation = 1
        difference = self.tol * 10000.0
        self.solution_history = np.zeros(self.max_generation + 1)
        self.solution_history[0] = np.min(self.parent_fitness)

        run_time = 0.0
        start_time = time.time()
        while converged is False and ngens < self.max_generation and \
                run_time < self.max_time:
            self.crossover()
            self.mutate()
            # determine fitness of offspring
            for i in range(self.population_size):
                self.chromosome_2_variables(self.offspring_population[i])
                self.offspring_fitness[i] = \
                    self.objective_function(self.design_variables)

            # rank the total population from best to worst
            total_fitness = np.append(self.parent_fitness,
                                      self.offspring_fitness)
            ranked_fitness = \
                np.argsort(total_fitness)[0:int(self.population_size)]

            total_population = \
                np.vstack([self.parent_population, self.offspring_population])
            self.parent_population[:, :] = total_population[ranked_fitness, :]
            self.parent_fitness[:] = total_fitness[ranked_fitness]

            # store solution history and wrap up generation
            self.solution_history[generation] = np.min(self.parent_fitness)

            if generation > self.convergence_iters:
                difference = \
                    self.solution_history[generation - self.convergence_iters]\
                    - self.solution_history[generation]
            else:
                difference = 1000
            if abs(difference) <= self.tol:
                converged = True

            # shuffle up the order of the population
            shuffle_order = np.arange(1, self.population_size)
            np.random.shuffle(shuffle_order)
            shuffle_order = np.append([0], shuffle_order)
            self.parent_population = self.parent_population[shuffle_order]
            self.parent_fitness = self.parent_fitness[shuffle_order]

            generation += 1
            ngens += 1

            run_time = time.time() - start_time

        # Assign final outputs
        self.solution_history = self.solution_history[0:ngens]
        self.optimized_function_value = np.min(self.parent_fitness)
        self.chromosome_2_variables(
            self.parent_population[np.argmin(self.parent_fitness)])
        self.optimized_design_variables = self.design_variables

        logger.debug('The GA ran for this many generations: {}'
                     .format(ngens))
        logger.debug('The GA ran for this many seconds: {:.3f}'
                     .format(run_time))
        logger.debug('The optimized function value was: {:.3e}'
                     .format(self.optimized_function_value))
        logger.debug('The optimal design variables were: {}'
                     .format(self.optimized_design_variables))



# -*- coding: utf-8 -*-
"""
File collection CLI utility functions.
"""
import logging

from gaps.cli.collect import collect
from gaps.cli import as_click_command, CLICommandFromFunction
from gaps.cli.preprocessing import preprocess_collect_config as _preprocessor

from reV.utilities import ModuleName


logger = logging.getLogger(__name__)


SPLIT_KEYS = [("_out_path", "_pattern")]
collect_command = CLICommandFromFunction(collect, name=str(ModuleName.COLLECT),
                                         split_keys=SPLIT_KEYS,
                                         config_preprocessor=_preprocessor)
main = as_click_command(collect_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Collect CLI.')
        raise



# -*- coding: utf-8 -*-
"""
General CLI utility functions.
"""
import logging
from warnings import warn

from gaps.pipeline import Status
from rex.utilities.loggers import init_mult

from reV.utilities import ModuleName
from reV.utilities.exceptions import ConfigWarning, PipelineError


logger = logging.getLogger(__name__)


def init_cli_logging(name, log_directory, verbose):
    """Initialize CLI logger

    Parameters
    ----------
    name : str
        The name to use for the log file written to disk.
    log_directory : str
        Path to log file output directory.
    verbose : bool
        Option to make logger verbose (DEBUG).
    """
    init_mult(name, log_directory, modules=['reV', 'rex'], verbose=verbose)
    logger.info("Initialized reV/rex {}loggers with name {!r} and log "
                "directory {!r}"
                .format("verbose " if verbose else "", name,
                        str(log_directory)))


def format_analysis_years(analysis_years=None):
    """Format user's analysis_years input

    Parameters
    ----------
    analysis_years : int | str | list, optional
        Years to run reV analysis on. Can be an integer or string, or a
        list of integers or strings (or ``None``). This input will get
        converted to a list of values automatically. If ``None``, a
        ``ConfigWarning`` will be thrown. By default, ``None``.

    Returns
    -------
    list
        List of analysis years. This list will never be empty, but it
        can contain ``None`` as the only value.
    """

    if not isinstance(analysis_years, list):
        analysis_years = [analysis_years]

    if analysis_years[0] is None:
        warn('Years may not have been specified, may default '
             'to available years in inputs files.', ConfigWarning)

    return analysis_years


def parse_from_pipeline(config, out_dir, config_key, target_modules):
    """Parse the out file from target modules and set as the values for key.

    This function only updates the ``config_key`` input if it is set to
    ``"PIPELINE"``.

    Parameters
    ----------
    config : dict
        Configuration dictionary. The ``config_key`` will be updated in
        this dictionary if it is set to ``"PIPELINE"``.
    out_dir : str
        Path to pipeline project directory where config and status files
        are located. The status file is expected to be in this
        directory.
    config_key : str
        Key in config files to replace with ``"out_file"`` value(s) from
        previous pipeline step.
    target_modules : list of str | list of `ModuleName`
        List of (previous) target modules to parse for the
        ``config_key``.

    Returns
    -------
    dict
        Input config dictionary with updated ``config_key`` input.

    Raises
    ------
    PipelineError
        If ``"out_file"`` not found in previous target module status
        files.
    """
    if config.get(config_key, None) == 'PIPELINE':
        for target_module in target_modules:
            gen_config_key = "gen" in config_key
            module_sca = target_module == ModuleName.SUPPLY_CURVE_AGGREGATION
            if gen_config_key and module_sca:
                target_key = "gen_fpath"
            else:
                target_key = "out_file"
            val = Status.parse_step_status(out_dir, target_module, target_key)
            if len(val) == 1:
                break
        else:
            raise PipelineError('Could not parse {} from previous '
                                'pipeline jobs.'.format(config_key))

        config[config_key] = val[0]
        logger.info('Config using the following pipeline input for {}: {}'
                    .format(config_key, val[0]))

    return config



# -*- coding: utf-8 -*-
"""Output request config to handle user output requests.

This module will allow for aliases and fix some typos.

Created on Mon Jul  8 09:37:23 2019

@author: gbuster
"""
import logging
from warnings import warn
from reV.utilities.exceptions import ConfigWarning


logger = logging.getLogger(__name__)


class OutputRequest(list):
    """Base output request list framework with request key correction logic."""

    # map of commonly expected typos.
    # keys are typos, values are correct var names
    # all available output variables should be in the values
    CORRECTIONS = {}

    def __init__(self, inp):
        """
        Parameters
        ----------
        inp : list | tuple | str
            List of requested reV output variables.
        """

        if isinstance(inp, str):
            inp = [inp]

        for request in inp:
            if request in self.CORRECTIONS.values():
                self.append(request)
            elif request in self.CORRECTIONS:
                self.append(self.CORRECTIONS[request])
                msg = ('Correcting output request "{}" to "{}".'
                       .format(request, self.CORRECTIONS[request]))
                logger.warning(msg)
                warn(msg, ConfigWarning)
            else:
                self.append(request)
                logger.debug('Did not recognize requested output variable '
                             '"{}". Passing forward, but this may cause a '
                             'downstream error. Available known output '
                             'variables are: {}'
                             .format(request,
                                     list(set(self.CORRECTIONS.values()))))


class SAMOutputRequest(OutputRequest):
    """SAM output request framework."""

    # map of commonly expected typos.
    # keys are typos, values are correct SAM var names
    # all available SAM output variables should be in the values
    CORRECTIONS = {'cf_means': 'cf_mean',
                   'cf': 'cf_mean',
                   'capacity': 'system_capacity',
                   'capacity_factor': 'cf_mean',
                   'capacityfactor': 'cf_mean',
                   'cf_profiles': 'cf_profile',
                   'profiles': 'cf_profile',
                   'profile': 'cf_profile',
                   'dni_means': 'dni_mean',
                   'ghi_means': 'ghi_mean',
                   'ws_means': 'ws_mean',
                   'generation': 'annual_energy',
                   'yield': 'energy_yield',
                   'generation_profile': 'gen_profile',
                   'generation_profiles': 'gen_profile',
                   'plane_of_array': 'poa',
                   'plane_of_array_irradiance': 'poa',
                   'gen_profiles': 'gen_profile',
                   'lcoe': 'lcoe_fcr',
                   'foc': 'fixed_operating_cost',
                   'voc': 'variable_operating_cost',
                   'fcr': 'fixed_charge_rate',
                   'cc': 'capital_cost',
                   'lcoe_nominal': 'lcoe_nom',
                   'real_lcoe': 'lcoe_real',
                   'net_present_value': 'project_return_aftertax_npv',
                   'npv': 'project_return_aftertax_npv',
                   'ppa': 'ppa_price',
                   'single_owner': 'ppa_price',
                   'singleowner': 'ppa_price',
                   'actual_irr': 'flip_actual_irr',
                   'irr': 'flip_actual_irr',
                   'cf_total_revenue': 'gross_revenue',
                   'total_cost': 'total_installed_cost',
                   'turbine': 'turbine_cost',
                   'sales_tax': 'sales_tax_cost',
                   'bos': 'bos_cost',
                   'albedo': 'surface_albedo',
                   'ac_power': 'ac',
                   'dc_power': 'dc',
                   'clipping': 'clipped_power',
                   'clipped': 'clipped_power',
                   'clip': 'clipped_power'
                   }



# -*- coding: utf-8 -*-
"""
Classes to collect reV outputs from multiple annual files.
"""
import glob
import logging
import os
import time
from warnings import warn

import numpy as np
import pandas as pd
from gaps.pipeline import parse_previous_status
from rex import Resource
from rex.utilities.utilities import (
    get_class_properties,
    get_lat_lon_cols,
    parse_year,
)

from reV.generation.base import LCOE_REQUIRED_OUTPUTS
from reV.config.output_request import SAMOutputRequest
from reV.handlers.outputs import Outputs
from reV.utilities import ModuleName, log_versions
from reV.utilities.exceptions import ConfigError, HandlerRuntimeError

logger = logging.getLogger(__name__)


class MultiYearGroup:
    """
    Handle group parameters
    """

    def __init__(self, name, out_dir, source_files=None,
                 source_dir=None, source_prefix=None,
                 source_pattern=None,
                 dsets=('cf_mean',), pass_through_dsets=None):
        """
        Parameters
        ----------
        name : str
            Group name. Can be ``"none"`` for no collection groups.
        out_dir : str
            Output directory - used for Pipeline handling.
        source_files : str | list, optional
            Explicit list of source files. Use either this input *OR*
            `source_dir` + `source_prefix`. If this input is
            ``"PIPELINE"``, the `source_files` input is determined from
            the status file of the previous pipeline step.
            If ``None``, use `source_dir` and `source_prefix`.
            By default, ``None``.
        source_dir : str, optional
            Directory to extract source files from (must be paired with
            `source_prefix`). By default, ``None``.
        source_prefix : str, optional
            File prefix to search for in source directory (must be
            paired with `source_dir`). By default, ``None``.
        source_pattern : str, optional
            Optional unix-style ``/filepath/pattern*.h5`` to specify the
            source files. This takes priority over `source_dir` and
            `source_prefix` but is not used if `source_files` are
            specified explicitly. By default, ``None``.
        dsets : str | list | tuple, optional
            List of datasets to collect. This can be set to
            ``"PIPELINE"`` if running from the command line as part of a
            reV pipeline. In this case, all the datasets from the
            previous pipeline step will be collected.
            By default, ``('cf_mean',)``.
        pass_through_dsets : list | tuple, optional
            Optional list of datasets that are identical in the
            multi-year files (e.g. input datasets that don't vary from
            year to year) that should be copied to the output multi-year
            file once without a year suffix or means/stdev calculation.
            By default, ``None``.
        """
        self._name = name
        self._dirout = out_dir
        self._source_files = source_files
        self._source_dir = source_dir
        self._source_prefix = source_prefix
        self._source_pattern = source_pattern
        self._pass_through_dsets = None
        self._dsets = None

        self._parse_pass_through_dsets(dsets, pass_through_dsets or [])
        self._parse_dsets(dsets)

    def _parse_pass_through_dsets(self, dsets, pass_through_dsets):
        """Parse a multi-year pass-through dataset collection request.

        Parameters
        ----------
        dsets : str | list
            One or more datasets to collect, or "PIPELINE"
        pass_through_dsets : list
            List of pass through datasets.
        """
        with Resource(self.source_files[0]) as res:
            all_dsets = res.datasets

        if isinstance(dsets, str) and dsets == 'PIPELINE':
            dsets = all_dsets

        if "lcoe_fcr" in dsets:
            for dset in LCOE_REQUIRED_OUTPUTS:
                if dset not in pass_through_dsets and dset in all_dsets:
                    pass_through_dsets.append(dset)

        if "dc_ac_ratio" in dsets:
            if "dc_ac_ratio" not in pass_through_dsets:
                pass_through_dsets.append("dc_ac_ratio")

        self._pass_through_dsets = SAMOutputRequest(pass_through_dsets)

    def _parse_dsets(self, dsets):
        """Parse a multi-year dataset collection request. Can handle PIPELINE
        argument which will find all datasets from one of the files being
        collected ignoring meta, time index, and pass_through_dsets

        Parameters
        ----------
        dsets : str | list
            One or more datasets to collect, or "PIPELINE"
        """
        if isinstance(dsets, str) and dsets == 'PIPELINE':
            files = parse_previous_status(self._dirout, ModuleName.MULTI_YEAR)
            with Resource(files[0]) as res:
                dsets = [d for d in res
                         if not d.startswith('time_index')
                         and d != 'meta'
                         and d not in self.pass_through_dsets]

        self._dsets = SAMOutputRequest(dsets)

    @property
    def name(self):
        """
        Returns
        -------
        name : str
            Group name
        """
        name = self._name if self._name.lower() != "none" else None
        return name

    @property
    def source_files(self):
        """
        Returns
        -------
        source_files : list
            list of source files to collect from
        """
        if self._source_files is not None:
            if isinstance(self._source_files, (list, tuple)):
                source_files = self._source_files
            elif self._source_files == "PIPELINE":
                source_files = parse_previous_status(self._dirout,
                                                     ModuleName.MULTI_YEAR)
            else:
                e = "source_files must be a list, tuple, or 'PIPELINE'"
                logger.error(e)
                raise ConfigError(e)

        elif self._source_pattern:
            source_files = glob.glob(self._source_pattern)
            if not all(fp.endswith('.h5') for fp in source_files):
                msg = ('Source pattern resulted in non-h5 files that cannot '
                       'be collected: {}, pattern: {}'
                       .format(source_files, self._source_pattern))
                logger.error(msg)
                raise RuntimeError(msg)

        elif self._source_dir and self._source_prefix:
            source_files = []
            for file in os.listdir(self._source_dir):
                if (file.startswith(self._source_prefix)
                        and file.endswith('.h5') and '_node' not in file):
                    source_files.append(os.path.join(self._source_dir,
                                                     file))
        else:
            e = ("source_files or both source_dir and "
                 "source_prefix must be provided")
            logger.error(e)
            raise ConfigError(e)

        if not any(source_files):
            e = ('Could not find any source files for '
                 'multi-year collection group: "{}" in "{}"'
                 .format(self.name, self._source_dir))
            logger.error(e)
            raise FileNotFoundError(e)

        return source_files

    @property
    def dsets(self):
        """
        Returns
        -------
        _dsets :list | tuple
            Datasets to collect
        """
        return self._dsets

    @property
    def pass_through_dsets(self):
        """Optional list of datasets that are identical in the multi-year
        files (e.g. input datasets that don't vary from year to year) that
        should be copied to the output multi-year file once without a
        year suffix or means/stdev calculation

        Returns
        -------
        list | tuple | None
        """
        return self._pass_through_dsets

    def _dict_rep(self):
        """Get a dictionary representation of this multi year collection group

        Returns
        -------
        dict
        """
        props = get_class_properties(self.__class__)
        out = {k: getattr(self, k) for k in props}
        out['group'] = self.name
        return out

    @classmethod
    def _factory(cls, out_dir, groups_dict):
        """
        Generate dictionary of MultiYearGroup objects for all groups in groups

        Parameters
        ----------
        out_dir : str
            Output directory, used for Pipeline handling
        groups_dict : dict
            Dictionary of group parameters, parsed from multi-year config file

        Returns
        -------
        groups : dict
            Dictionary of MultiYearGroup objects for each group in groups
        """
        groups = {}
        for name, kwargs in groups_dict.items():
            groups[name] = cls(name, out_dir, **kwargs)

        return groups


class MultiYear(Outputs):
    """
    Class to handle multiple years of data and:
    - collect datasets from multiple years
    - compute multi-year means
    - compute multi-year standard deviations
    - compute multi-year coefficient of variations

    """

    def __init__(self, h5_file, group=None, unscale=True, mode='r',
                 str_decode=True):
        """
        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        group : str
            Group to collect datasets into
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        mode : str
            Mode to instantiate h5py.File instance
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        """
        log_versions(logger)
        super().__init__(h5_file, group=group, unscale=unscale, mode=mode,
                         str_decode=str_decode)

    @staticmethod
    def _create_dset_name(source_h5, dset):
        """
        Create output dataset name by parsing year from source_h5 and
        appending to source dataset name.

        Parameters
        ----------
        source_h5 : str
            Path to source .h5 file to copy data from
        dset : str
            Dataset to copy

        Returns
        -------
        dset_out : str
            Ouput dataset name
        """
        f_name = os.path.basename(source_h5)
        year = parse_year(f_name)
        dset_out = "{}-{}".format(dset, year)
        return dset_out

    def _copy_time_index(self, source_h5):
        """
        Copy time_index from source_h5 to time_index-{year} in multiyear .h5

        Parameters
        ----------
        source_h5 : str
            Path to source .h5 file to copy data from
        """
        dset_out = self._create_dset_name(source_h5, 'time_index')
        if dset_out not in self.datasets:
            logger.debug("- Collecting time_index from {}"
                         .format(os.path.basename(source_h5)))
            with Outputs(source_h5, mode='r') as f_in:
                time_index = f_in.h5['time_index'][...]

            self._create_dset(dset_out, time_index.shape, time_index.dtype,
                              data=time_index)

    def _copy_dset(self, source_h5, dset, meta=None, pass_through=False):
        """
        Copy dset_in from source_h5 to multiyear .h5

        Parameters
        ----------
        source_h5 : str
            Path to source .h5 file to copy data from
        dset : str
            Dataset to copy
        meta : pandas.DataFrame
            If provided confirm that source meta matches given meta
        pass_through : bool
            Flag to just pass through dataset without name modifications
            (no differences between years, no means or stdevs)
        """
        if pass_through:
            dset_out = dset
        else:
            dset_out = self._create_dset_name(source_h5, dset)

        if dset_out not in self.datasets:
            logger.debug("- Collecting {} from {}"
                         .format(dset, os.path.basename(source_h5)))
            with Outputs(source_h5, unscale=False, mode='r') as f_in:
                if meta is not None:
                    cols = get_lat_lon_cols(meta)
                    source_meta = f_in.meta

                    if len(meta) != len(source_meta):
                        msg = ('Meta data has different lengths between '
                               'collection files! Found {} and {}'
                               .format(len(meta), len(source_meta)))
                        logger.error(msg)
                        raise HandlerRuntimeError(msg)

                    if not np.allclose(meta[cols], source_meta[cols]):
                        msg = ('Coordinates do not match between '
                               'collection files!')
                        logger.warning(msg)
                        warn(msg)

                _, ds_dtype, ds_chunks = f_in.get_dset_properties(dset)
                ds_attrs = f_in.get_attrs(dset=dset)
                ds_data = f_in[dset]

            self._create_dset(dset_out, ds_data.shape, ds_dtype,
                              chunks=ds_chunks, attrs=ds_attrs, data=ds_data)

    @staticmethod
    def parse_source_files_pattern(source_files):
        """Parse a source_files pattern that can be either an explicit list of
        source files or a unix-style /filepath/pattern*.h5 and either way
        return a list of explicit filepaths.

        Parameters
        ----------
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to

        Returns
        -------
        source_files : list
            List of .h5 filepaths.
        """

        if isinstance(source_files, str) and '*' in source_files:
            source_files = glob.glob(source_files)
        elif isinstance(source_files, str):
            source_files = [source_files]
        elif not isinstance(source_files, (list, tuple)):
            msg = ('Cannot recognize source_files type: {} {}'
                   .format(source_files, type(source_files)))
            logger.error(msg)
            raise TypeError(msg)

        if not all(fp.endswith('.h5') for fp in source_files):
            msg = ('Non-h5 files cannot be collected: {}'.format(source_files))
            logger.error(msg)
            raise RuntimeError(msg)

        return source_files

    def collect(self, source_files, dset, profiles=False, pass_through=False):
        """
        Collect dataset dset from given list of h5 files

        Parameters
        ----------
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to
        dset : str
            Dataset to collect
        profiles : bool
            Boolean flag to indicate if profiles are being collected
            If True also collect time_index
        pass_through : bool
            Flag to just pass through dataset without name modifications
            (no differences between years, no means or stdevs)
        """
        source_files = self.parse_source_files_pattern(source_files)
        with Outputs(source_files[0], mode='r') as f_in:
            meta = f_in.h5['meta'][...]

        if 'meta' not in self.datasets:
            logger.debug("Copying meta")
            self._create_dset('meta', meta.shape, meta.dtype,
                              data=meta)

        meta = pd.DataFrame(meta)
        for year_h5 in source_files:
            if profiles:
                self._copy_time_index(year_h5)

            self._copy_dset(year_h5, dset, meta=meta,
                            pass_through=pass_through)

    def _get_source_dsets(self, dset_out):
        """
        Extract all available annual datasets associated with dset

        Parameters
        ----------
        dset_out : str
            Output dataset to find source datasets for

        Returns
        -------
        source_dsets : list
            List of annual datasets
        """
        dset = os.path.basename(dset_out).split("-")[0]
        logger.debug('-- source_dset root = {}'.format(dset))
        my_dset = ["{}-{}".format(dset, val) for val in ['means', 'stdev']]
        source_dsets = [ds for ds in self.datasets if dset in ds
                        and ds not in my_dset]
        if dset_out in source_dsets:
            source_dsets.remove(dset_out)

        return source_dsets

    def _update_dset(self, dset_out, dset_data):
        """
        Update dataset, create if needed

        Parameters
        ----------
        dset_out : str
            Dataset name
        dset_data : ndarray
            Dataset data to write to disc
        """
        if dset_out in self.datasets:
            logger.debug("- Updating {}".format(dset_out))
            self[dset_out] = dset_data
        else:
            logger.debug("- Creating {}".format(dset_out))
            source_dset = self._get_source_dsets(dset_out)[0]
            _, ds_dtype, ds_chunks = self.get_dset_properties(source_dset)
            ds_attrs = self.get_attrs(dset=source_dset)
            self._add_dset(dset_out, dset_data, ds_dtype,
                           chunks=ds_chunks, attrs=ds_attrs)

    def _compute_means(self, dset_out):
        """
        Compute multi-year means for given dataset

        Parameters
        ----------
        dset_out : str
            Multi-year means dataset name

        Returns
        -------
        my_means : ndarray
            Array of multi-year means
        """
        source_dsets = self._get_source_dsets(dset_out)
        logger.debug('\t- Computing {} from {}'.format(dset_out, source_dsets))

        my_means = np.zeros(len(self), dtype='float32')
        for ds in source_dsets:
            if self.h5[ds].shape == my_means.shape:
                my_means += self[ds]
            else:
                raise HandlerRuntimeError("{} shape {} should be {}"
                                          .format(ds, self.h5[ds].shape,
                                                  my_means.shape))
        my_means /= len(source_dsets)
        self._update_dset(dset_out, my_means)

        return my_means

    def means(self, dset):
        """
        Extract or compute multi-year means for given source dset

        Parameters
        ----------
        dset : str
            Dataset of interest

        Returns
        -------
        my_means : ndarray
            Array of multi-year means for dataset of interest
        """
        my_dset = "{}-means".format(dset)
        if my_dset in self.datasets:
            my_means = self[my_dset]
        else:
            my_means = self._compute_means(my_dset)

        return my_means

    def _compute_stdev(self, dset_out, means=None):
        """
        Compute multi-year standard deviation for given dataset

        Parameters
        ----------
        dset_out : str
            Multi-year stdev dataset name
        means : ndarray
            Array of pre-computed means

        Returns
        -------
        my_stdev : ndarray
            Array of multi-year standard deviations
        """
        if means is None:
            means = self._compute_means("{}-means".format(dset_out))

        source_dsets = self._get_source_dsets(dset_out)

        my_stdev = np.zeros(means.shape, dtype='float32')
        for ds in source_dsets:
            if self.h5[ds].shape == my_stdev.shape:
                my_stdev += (self[ds] - means)**2
            else:
                raise HandlerRuntimeError("{} shape {} should be {}"
                                          .format(ds, self.h5[ds].shape,
                                                  my_stdev.shape))

        my_stdev = np.sqrt(my_stdev / len(source_dsets))
        self._update_dset(dset_out, my_stdev)

        return my_stdev

    def stdev(self, dset):
        """
        Extract or compute multi-year standard deviation for given source dset

        Parameters
        ----------
        dset : str
            Dataset of interest

        Returns
        -------
        my_stdev : ndarray
            Array of multi-year standard deviation for dataset of interest
        """
        my_dset = "{}-stdev".format(dset)
        if my_dset in self.datasets:
            my_stdev = self[my_dset]
        else:
            my_means = self.means(dset)
            my_stdev = self._compute_stdev(my_dset, means=my_means)

        return my_stdev

    def CV(self, dset):
        """
        Extract or compute multi-year coefficient of variation for given
        source dset

        Parameters
        ----------
        dset : str
            Dataset of interest

        Returns
        -------
        my_cv : ndarray
            Array of multi-year coefficient of variation for
            dataset of interest
        """
        my_cv = self.stdev(dset) / self.means(dset)
        return my_cv

    @classmethod
    def is_profile(cls, source_files, dset):
        """
        Check dataset in source files to see if it is a profile.

        Parameters
        ----------
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to
        dset : str
            Dataset to collect

        Returns
        -------
        is_profile : bool
            True if profile, False if not.
        """
        source_files = cls.parse_source_files_pattern(source_files)
        with Outputs(source_files[0]) as f:
            if dset not in f.datasets:
                raise KeyError('Dataset "{}" not found in source file: "{}"'
                               .format(dset, source_files[0]))

            shape, _, _ = f.get_dset_properties(dset)

        return len(shape) == 2

    @classmethod
    def pass_through(cls, my_file, source_files, dset, group=None):
        """
        Pass through a dataset that is identical in all source files to a
        dataset of the same name in the output multi-year file.

        Parameters
        ----------
        my_file : str
            Path to multi-year .h5 file
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to
        dset : str
            Dataset to pass through (will also be the name of the output
            dataset in my_file)
        group : str
            Group to collect datasets into
        """
        source_files = cls.parse_source_files_pattern(source_files)
        logger.info('Passing through {} into {}.'
                    .format(dset, my_file))
        with cls(my_file, mode='a', group=group) as my:
            my.collect(source_files, dset, pass_through=True)

    @classmethod
    def collect_means(cls, my_file, source_files, dset, group=None):
        """
        Collect and compute multi-year means for given dataset

        Parameters
        ----------
        my_file : str
            Path to multi-year .h5 file
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to
        dset : str
            Dataset to collect
        group : str
            Group to collect datasets into
        """
        logger.info('Collecting {} into {} '
                    'and computing multi-year means and standard deviations.'
                    .format(dset, my_file))
        source_files = cls.parse_source_files_pattern(source_files)
        with cls(my_file, mode='a', group=group) as my:
            my.collect(source_files, dset)
            means = my._compute_means("{}-means".format(dset))
            my._compute_stdev("{}-stdev".format(dset), means=means)

    @classmethod
    def collect_profiles(cls, my_file, source_files, dset, group=None):
        """
        Collect multi-year profiles associated with given dataset

        Parameters
        ----------
        my_file : str
            Path to multi-year .h5 file
        source_files : list | str
            List of .h5 files to collect datasets from. This can also be a
            unix-style /filepath/pattern*.h5 to find .h5 files to collect,
            however all resulting files must be .h5 otherwise an exception will
            be raised. NOTE: .h5 file names must indicate the year the data
            pertains to
        dset : str
            Profiles dataset to collect
        group : str
            Group to collect datasets into
        """
        logger.info('Collecting {} into {}'.format(dset, my_file))
        source_files = cls.parse_source_files_pattern(source_files)
        with cls(my_file, mode='a', group=group) as my:
            my.collect(source_files, dset, profiles=True)


def my_collect_groups(out_fpath, groups, clobber=True):
    """Collect all groups into a single multi-year HDF5 file.

    ``reV`` multi-year combines ``reV`` generation data from multiple
    years (typically stored in separate files) into a single multi-year
    file. Each dataset in the multi-year file is labeled with the
    corresponding years, and multi-year averages of the yearly datasets
    are also computed.

    Parameters
    ----------
    out_fpath : str
        Path to multi-year HDF5 file to use for multi-year
        collection.
    groups : dict
        Dictionary of collection groups and their parameters. This
        should be a dictionary mapping group names (keys) to a set
        of key word arguments (values) that can be used to initialize
        :class:`~reV.handlers.multi_year.MultiYearGroup` (excluding the
        required ``name`` and ``out_dir`` inputs, which are populated
        automatically). For example::

            groups = {
                "none": {
                    "dsets": [
                        "cf_profile",
                        "cf_mean",
                        "ghi_mean",
                        "lcoe_fcr",
                    ],
                    "source_dir": "./",
                    "source_prefix": "",
                    "pass_through_dsets": [
                        "capital_cost",
                        "fixed_operating_cost",
                        "system_capacity",
                        "fixed_charge_rate",
                        "variable_operating_cost",
                    ]
                },
                "solar_group": {
                    "source_files": "PIPELINE",
                    "dsets": [
                        "cf_profile_ac",
                        "cf_mean_ac",
                        "ac",
                        "dc",
                        "clipped_power"
                    ],
                    "pass_through_dsets": [
                        "system_capacity_ac",
                        "dc_ac_ratio"
                    ]
                },
                ...
            }

        The group names will be used as the HDF5 file group name under
        which the collected data will be stored. You can have exactly
        one group with the name ``"none"`` for a "no group" collection
        (this is typically what you want and all you need to specify).
    clobber : bool, optional
        Flag to purge the multi-year output file prior to running the
        multi-year collection step if the file already exists on disk.
        This ensures the data is always freshly collected from the
        single-year files. If ``False``, then datasets in the existing
        file will **not** be overwritten with (potentially new/updated)
        data from the single-year files. By default, ``True``.
    """
    if not out_fpath.endswith(".h5"):
        out_fpath = '{}.h5'.format(out_fpath)

    if clobber and os.path.exists(out_fpath):
        msg = ('Found existing multi-year file: "{}". Removing...'
               .format(str(out_fpath)))
        logger.warning(msg)
        warn(msg)
        os.remove(out_fpath)

    out_dir = os.path.dirname(out_fpath)
    groups = MultiYearGroup._factory(out_dir, groups)
    group_params = {name: group._dict_rep()
                    for name, group in groups.items()}

    logger.info('Multi-year collection is being run with output path: {}'
                .format(out_fpath))
    ts = time.time()
    for group_name, group in group_params.items():
        logger.info('- Collecting datasets "{}" from "{}" into "{}/"'
                    .format(group['dsets'], group['source_files'],
                            group_name))
        t0 = time.time()
        for dset in group['dsets']:
            if MultiYear.is_profile(group['source_files'], dset):
                MultiYear.collect_profiles(out_fpath, group['source_files'],
                                           dset, group=group['group'])
            else:
                MultiYear.collect_means(out_fpath, group['source_files'],
                                        dset, group=group['group'])

        pass_through_dsets = group.get('pass_through_dsets') or []
        for dset in pass_through_dsets:
            MultiYear.pass_through(out_fpath, group['source_files'],
                                   dset, group=group['group'])

        runtime = (time.time() - t0) / 60
        logger.info('- {} collection completed in: {:.2f} min.'
                    .format(group_name, runtime))

    runtime = (time.time() - ts) / 60
    logger.info('Multi-year collection completed in : {:.2f} min.'
                .format(runtime))

    return out_fpath



# -*- coding: utf-8 -*-
"""
reV Representative Profiles CLI utility functions.
"""
import os
import glob
import logging
from warnings import warn

from rex.utilities.utilities import parse_year
from gaps.cli import as_click_command, CLICommandFromClass

from reV.hybrids.hybrids import Hybridization
from reV.utilities.exceptions import PipelineError
from reV.utilities import ModuleName


logger = logging.getLogger(__name__)


def _preprocessor(config, out_dir, job_name):
    """Preprocess hybrids config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    out_dir : str
        Path to output file directory.
    job_name : str
        Name of hybrids job. This will be included in the output
        file name.

    Returns
    -------
    dict
        Updated config file.
    """
    for key in ["solar_fpath", "wind_fpath"]:
        config[key] = _glob_to_yearly_dict(config[key])

    config = _set_paths(config, out_dir, job_name)
    return config


def _glob_to_yearly_dict(fpath):
    """Glob the filepaths into a dictionary based on years. """
    _raise_err_if_pipeline(fpath)
    paths = {}
    for fp in glob.glob(fpath):
        fname = os.path.basename(fp)

        try:
            year = parse_year(fname)
        except RuntimeError:
            year = None

        paths.setdefault(year, []).append(fp)

    return paths


def _raise_err_if_pipeline(fpath):
    """Raise error if fpath input is 'PIPELINE'. """

    if fpath == 'PIPELINE':
        msg = ('Hybrids module cannot infer fpath from "PIPELINE" - '
               'input is ambiguous. Please specify both the solar and '
               'wind fpath before running hybrids module.')
        logger.error(msg)
        raise PipelineError(msg)


def _set_paths(config, out_dir, job_name):
    """Pair solar and wind files and corresponding process names. """

    solar_glob_paths = config["solar_fpath"]
    wind_glob_paths = config["wind_fpath"]
    all_years = set(solar_glob_paths) | set(wind_glob_paths)
    common_years = set(solar_glob_paths) & set(wind_glob_paths)
    if not all_years:
        msg = "No files found that match the input: {!r} and/or {!r}"
        e = msg.format(config['solar_fpath'], config['wind_fpath'])
        logger.error(e)
        raise RuntimeError(e)

    solar_fpaths = []
    wind_fpaths = []
    out_files = []
    for year in all_years:
        if year not in common_years:
            msg = ("No corresponding {} file found for {} input file "
                   "(year: '{}'): {!r}. No hybridization performed for "
                   "this input!")
            resources = (['solar', 'wind'] if year not in solar_glob_paths else
                         ['wind', 'solar'])
            paths = (solar_glob_paths.get(year, [])
                     + wind_glob_paths.get(year, []))
            w = msg.format(*resources, paths, year)
            logger.warning(w)
            warn(w)
            continue

        for fpaths in (solar_glob_paths, wind_glob_paths):
            if len(fpaths[year]) > 1:
                msg = ("Ambiguous number of files found for year '{}': {!r} "
                       "Please ensure there is only one input file per year. "
                       "No hybridization performed for this input!")
                w = msg.format(year, fpaths[year])
                logger.warning(w)
                warn(w)
                break
        else:
            solar_fpaths += solar_glob_paths[year]
            wind_fpaths += wind_glob_paths[year]
            out_fn = ("{}.h5".format(job_name)
                      if year is None
                      else "{}_{}.h5".format(job_name, year))
            out_files += [os.path.join(out_dir, out_fn)]

    if not solar_fpaths or not wind_fpaths:
        msg = "No files found that match the input: {!r} and/or {!r}"
        e = msg.format(config['solar_fpath'], config['wind_fpath'])
        logger.error(e)
        raise RuntimeError(e)

    config["solar_fpath"] = solar_fpaths
    config["wind_fpath"] = wind_fpaths
    config["fout"] = out_files

    return config


SPLIT_KEYS = [("solar_fpath", "wind_fpath", "fout")]
hybrids_command = CLICommandFromClass(Hybridization, method="run",
                                      name=str(ModuleName.HYBRIDS),
                                      add_collect=False, split_keys=SPLIT_KEYS,
                                      config_preprocessor=_preprocessor,
                                      skip_doc_params=["fout"])
main = as_click_command(hybrids_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Hybrids CLI.')
        raise



# -*- coding: utf-8 -*-
"""reV-to-SAM econ interface module.

Wraps the NREL-PySAM lcoefcr and singleowner modules with
additional reV features.
"""
import logging
from copy import deepcopy
from warnings import warn

import numpy as np
import PySAM.Lcoefcr as PySamLCOE
import PySAM.Singleowner as PySamSingleOwner

from reV.handlers.outputs import Outputs
from reV.SAM.defaults import DefaultLCOE, DefaultSingleOwner
from reV.SAM.SAM import RevPySam
from reV.SAM.windbos import WindBos
from reV.utilities.exceptions import SAMExecutionError
from reV.utilities import ResourceMetaField

logger = logging.getLogger(__name__)


class Economic(RevPySam):
    """Base class for SAM economic models."""

    MODULE = None

    def __init__(self, sam_sys_inputs, site_sys_inputs=None,
                 output_request='lcoe_fcr'):
        """Initialize a SAM economic model object.

        Parameters
        ----------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list | tuple | str
            Requested SAM output(s) (e.g., 'ppa_price', 'lcoe_fcr').
        """

        self._site = None

        if isinstance(output_request, (list, tuple)):
            self.output_request = output_request
        else:
            self.output_request = (output_request,)

        super().__init__(meta=None, sam_sys_inputs=sam_sys_inputs,
                         site_sys_inputs=site_sys_inputs,
                         output_request=output_request)

    @staticmethod
    def _parse_sys_cap(site, inputs, site_df):
        """Find the system capacity variable in either inputs or df.

        Parameters
        ----------
        site : int
            Site gid.
        inputs : dict
            Generic system inputs (not site-specific).
        site_df : pd.DataFrame
            Site-specific inputs table with index = site gid's

        Returns
        -------
        sys_cap : int | float
            System nameplate capacity in native units (SAM is kW).
        """

        if ('system_capacity' not in inputs
                and 'turbine_capacity' not in inputs
                and 'system_capacity' not in site_df
                and 'turbine_capacity' not in site_df):
            raise SAMExecutionError('Input parameter "system_capacity" '
                                    'or "turbine_capacity" '
                                    'must be included in the SAM config '
                                    'inputs or site-specific inputs in '
                                    'order to calculate annual energy '
                                    'yield for LCOE. Received the following '
                                    'inputs, site_df:\n{}\n{}'
                                    .format(inputs, site_df.head()))

        if 'system_capacity' in inputs:
            sys_cap = inputs['system_capacity']
        elif 'turbine_capacity' in inputs:
            sys_cap = inputs['turbine_capacity']
        elif 'system_capacity' in site_df:
            sys_cap = site_df.loc[site, 'system_capacity']
        elif 'turbine_capacity' in site_df:
            sys_cap = site_df.loc[site, 'turbine_capacity']

        return sys_cap

    @classmethod
    def _get_annual_energy(cls, site, site_df, site_gids, cf_arr, inputs,
                           calc_aey):
        """Get the single-site cf and annual energy and add to site_df.

        Parameters
        ----------
        site : int
            Site gid.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        site_gids : list
            List of all site gid values from the cf_file.
        cf_arr : np.ndarray
            Array of cf_mean values for all sites in the cf_file for the
            given year.
        inputs : dict
            Dictionary of SAM input parameters.
        calc_aey : bool
            Flag to add annual_energy to df.

        Returns
        -------
        site_df : pd.DataFrame
            Same as input but with added labels "capacity_factor" and
            "annual_energy" (latter is dependent on calc_aey flag).
        """

        # get the index location of the site in question
        isite = site_gids.index(site)

        # calculate the capacity factor
        cf = cf_arr[isite]
        if cf > 1:
            warn('Capacity factor > 1. Dividing by 100.')
            cf /= 100
        site_df.loc[site, 'capacity_factor'] = cf

        # calculate the annual energy yield if not input;
        if calc_aey:
            # get the system capacity
            sys_cap = cls._parse_sys_cap(site, inputs, site_df)

            # Calc annual energy, mult by 8760 to convert kW to kWh
            aey = sys_cap * cf * 8760

            # add aey to site-specific inputs
            site_df.loc[site, 'annual_energy'] = aey

        return site_df

    @staticmethod
    def _get_cf_profiles(sites, cf_file, year):
        """Get the multi-site capacity factor time series profiles.

        Parameters
        ----------
        sites : list
            List of all site GID's to get gen profiles for.
        cf_file : str
            reV generation capacity factor output file with path.
        year : int | str | None
            reV generation year to calculate econ for. Looks for cf_mean_{year}
            or cf_profile_{year}. None will default to a non-year-specific cf
            dataset (cf_mean, cf_profile).

        Returns
        -------
        profiles : np.ndarray
            2D array (time, n_sites) of all capacity factor profiles for all
            the requested sites.
        """

        # Retrieve the generation profile for single owner input
        with Outputs(cf_file) as cfh:

            # get the index location of the site in question
            site_gids = list(cfh.get_meta_arr(ResourceMetaField.GID))
            isites = [site_gids.index(s) for s in sites]

            # look for the cf_profile dataset
            if 'cf_profile' in cfh.datasets:
                dset = 'cf_profile'
            elif 'cf_profile-{}'.format(year) in cfh.datasets:
                dset = 'cf_profile-{}'.format(year)
            elif 'cf_profile_{}'.format(year) in cfh.datasets:
                dset = 'cf_profile_{}'.format(year)
            else:
                msg = ('Could not find cf_profile values for '
                       'input to SingleOwner. Available datasets: {}'
                       .format(cfh.datasets))
                logger.error(msg)
                raise KeyError(msg)

            profiles = cfh[dset, :, isites]

        return profiles

    @classmethod
    def _make_gen_profile(cls, isite, site, profiles, site_df, inputs):
        """Get the single-site generation time series and add to inputs dict.

        Parameters
        ----------
        isite : int
            Site index in the profiles array.
        site : int
            Site resource GID.
        profiles : np.ndarray
            2D array (time, n_sites) of all capacity factor profiles for all
            the requested sites.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        inputs : dict
            Dictionary of SAM input parameters.

        Returns
        -------
        inputs : dict
            Dictionary of SAM input parameters with the generation profile
            added.
        """

        sys_cap = cls._parse_sys_cap(site, inputs, site_df)
        inputs['gen'] = profiles[:, isite] * sys_cap

        return inputs

    def ppa_price(self):
        """Get PPA price ($/MWh).

        Native units are cents/kWh, mult by 10 for $/MWh.
        """
        return self['ppa'] * 10

    def npv(self):
        """Get net present value (NPV) ($).

        Native units are dollars.
        """
        return self['project_return_aftertax_npv']

    def lcoe_fcr(self):
        """Get LCOE ($/MWh).

        Native units are $/kWh, mult by 1000 for $/MWh.
        """
        if 'lcoe_fcr' in self.outputs:
            lcoe = self.outputs['lcoe_fcr']
        else:
            lcoe = self['lcoe_fcr'] * 1000
        return lcoe

    def lcoe_nom(self):
        """Get nominal LCOE ($/MWh) (from PPA/SingleOwner model).

        Native units are cents/kWh, mult by 10 for $/MWh.
        """
        return self['lcoe_nom'] * 10

    def lcoe_real(self):
        """Get real LCOE ($/MWh) (from PPA/SingleOwner model).

        Native units are cents/kWh, mult by 10 for $/MWh.
        """
        return self['lcoe_real'] * 10

    def flip_actual_irr(self):
        """Get actual IRR (from PPA/SingleOwner model).

        Native units are %.
        """
        return self['flip_actual_irr']

    def gross_revenue(self):
        """Get cash flow total revenue (from PPA/SingleOwner model).

        Native units are $.
        """
        cf_tr = np.array(self['cf_total_revenue'], dtype=np.float32)
        cf_tr = np.sum(cf_tr, axis=0)
        return cf_tr

    def collect_outputs(self):
        """Collect SAM output_request, convert timeseries outputs to UTC, and
        save outputs to self.outputs property.
        """

        output_lookup = {'ppa_price': self.ppa_price,
                         'project_return_aftertax_npv': self.npv,
                         'lcoe_fcr': self.lcoe_fcr,
                         'lcoe_nom': self.lcoe_nom,
                         'lcoe_real': self.lcoe_real,
                         'flip_actual_irr': self.flip_actual_irr,
                         'gross_revenue': self.gross_revenue,
                         }

        super().collect_outputs(output_lookup)

    @classmethod
    def reV_run(cls, site, site_df, inputs, output_request):
        """Run the SAM econ model for a single site.

        Parameters
        ----------
        site : int
            Site gid.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        inputs : dict
            Dictionary of SAM system input parameters.
        output_request : list | tuple | str
            Requested SAM output(s) (e.g., 'ppa_price', 'lcoe_fcr').

        Returns
        -------
        sim.outputs : dict
            Dictionary keyed by SAM variable names with SAM numerical results.
        """

        # Create SAM econ instance and calculate requested output.
        sim = cls(sam_sys_inputs=inputs,
                  site_sys_inputs=dict(site_df.loc[site, :]),
                  output_request=output_request)
        sim._site = site

        sim.assign_inputs()
        sim.execute()
        sim.collect_outputs()

        return sim.outputs


class LCOE(Economic):
    """SAM LCOE model.
    """

    MODULE = 'lcoefcr'
    PYSAM = PySamLCOE

    def __init__(self, sam_sys_inputs, site_sys_inputs=None,
                 output_request=('lcoe_fcr',)):
        """Initialize a SAM LCOE economic model object."""
        super().__init__(sam_sys_inputs, site_sys_inputs=site_sys_inputs,
                         output_request=output_request)

    @staticmethod
    def _parse_lcoe_inputs(site_df, cf_file, year):
        """Parse for non-site-specific LCOE inputs.

        Parameters
        ----------
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        cf_file : str
            reV generation capacity factor output file with path.
        year : int | str | None
            reV generation year to calculate econ for. Looks for cf_mean_{year}
            or cf_profile_{year}. None will default to a non-year-specific cf
            dataset (cf_mean, cf_profile).

        Returns
        -------
        site_gids : list
            List of all site gid values from the cf_file.
        calc_aey : bool
            Flag to require calculation of the annual energy yield before
            running LCOE.
        cf_arr : np.ndarray
            Array of cf_mean values for all sites in the cf_file for the
            given year.
        """

        # get the cf_file meta data gid's to use as indexing tools
        with Outputs(cf_file) as cfh:
            site_gids = list(cfh.meta[ResourceMetaField.GID])

        calc_aey = False
        if 'annual_energy' not in site_df:
            # annual energy yield has not been input, flag to calculate
            site_df.loc[:, 'annual_energy'] = np.nan
            calc_aey = True

        # make sure capacity factor is present in site-specific data
        if 'capacity_factor' not in site_df:
            site_df.loc[:, 'capacity_factor'] = np.nan

        # pull all cf mean values for LCOE calc
        with Outputs(cf_file) as cfh:
            if 'cf_mean' in cfh.datasets:
                cf_arr = cfh['cf_mean']
            elif 'cf_mean-{}'.format(year) in cfh.datasets:
                cf_arr = cfh['cf_mean-{}'.format(year)]
            elif 'cf_mean_{}'.format(year) in cfh.datasets:
                cf_arr = cfh['cf_mean_{}'.format(year)]
            elif 'cf' in cfh.datasets:
                cf_arr = cfh['cf']
            else:
                raise KeyError('Could not find cf_mean values for LCOE. '
                               'Available datasets: {}'.format(cfh.datasets))
        return site_gids, calc_aey, cf_arr

    @staticmethod
    def default():
        """Get the executed default pysam LCOE FCR object.

        Returns
        -------
        PySAM.Lcoefcr
        """
        return DefaultLCOE.default()

    @classmethod
    def reV_run(cls, points_control, site_df, cf_file, year,
                output_request=('lcoe_fcr',)):
        """Execute SAM LCOE simulations based on a reV points control instance.

        Parameters
        ----------
        points_control : config.PointsControl
            PointsControl instance containing project points site and SAM
            config info.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        cf_file : str
            reV generation capacity factor output file with path.
        year : int | str | None
            reV generation year to calculate econ for. Looks for cf_mean_{year}
            or cf_profile_{year}. None will default to a non-year-specific cf
            dataset (cf_mean, cf_profile).
        output_request : list | tuple | str
            Output(s) to retrieve from SAM.

        Returns
        -------
        out : dict
            Nested dictionaries where the top level key is the site index,
            the second level key is the variable name, second level value is
            the output variable value.
        """

        out = {}

        site_gids, calc_aey, cf_arr = cls._parse_lcoe_inputs(site_df, cf_file,
                                                             year)

        for site in points_control.sites:
            # get SAM inputs from project_points based on the current site
            _, inputs = points_control.project_points[site]

            site_df = cls._get_annual_energy(site, site_df, site_gids, cf_arr,
                                             inputs, calc_aey)

            out[site] = super().reV_run(site, site_df, inputs, output_request)

        return out


class SingleOwner(Economic):
    """SAM single owner economic model.
    """

    MODULE = 'singleowner'
    PYSAM = PySamSingleOwner

    def __init__(self, sam_sys_inputs, site_sys_inputs=None,
                 output_request=('ppa_price',)):
        """Initialize a SAM single owner economic model object.
        """
        super().__init__(sam_sys_inputs, site_sys_inputs=site_sys_inputs,
                         output_request=output_request)

        # run balance of system cost model if required
        self.sam_sys_inputs, self.windbos_outputs = \
            self._windbos(self.sam_sys_inputs)

    @staticmethod
    def _windbos(inputs):
        """Run SAM Wind Balance of System cost model if requested.

        Parameters
        ----------
        inputs : dict
            Dictionary of SAM key-value pair inputs.
            "total_installed_cost": "windbos" will trigger the windbos method.

        Returns
        -------
        inputs : dict
            Dictionary of SAM key-value pair inputs with the total installed
            cost replaced with WindBOS values if requested.
        output : dict
            Dictionary of windbos cost breakdowns.
        """

        outputs = {}
        if (inputs is not None
                and 'total_installed_cost' in inputs
                and isinstance(inputs['total_installed_cost'], str)
                and inputs['total_installed_cost'].lower() == 'windbos'):
            wb = WindBos(inputs)
            inputs['total_installed_cost'] = wb.total_installed_cost
            outputs = wb.output
        return inputs, outputs

    @staticmethod
    def default():
        """Get the executed default pysam Single Owner object.

        Returns
        -------
        PySAM.Singleowner
        """
        return DefaultSingleOwner.default()

    def collect_outputs(self):
        """Collect SAM output_request, convert timeseries outputs to UTC, and
        save outputs to self.outputs property. This includes windbos outputs.
        """

        windbos_out_vars = [v for v in self.output_request
                            if v in self.windbos_outputs]
        self.output_request = [v for v in self.output_request
                               if v not in windbos_out_vars]

        super().collect_outputs()

        windbos_results = {}
        for request in windbos_out_vars:
            windbos_results[request] = self.windbos_outputs[request]

        self.outputs.update(windbos_results)

    @classmethod
    def reV_run(cls, points_control, site_df, cf_file, year,
                output_request=('ppa_price',)):
        """Execute SAM SingleOwner simulations based on reV points control.

        Parameters
        ----------
        points_control : config.PointsControl
            PointsControl instance containing project points site and SAM
            config info.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        cf_file : str
            reV generation capacity factor output file with path.
        year : int | str | None
            reV generation year to calculate econ for. Looks for cf_mean_{year}
            or cf_profile_{year}. None will default to a non-year-specific cf
            dataset (cf_mean, cf_profile).
        output_request : list | tuple | str
            Output(s) to retrieve from SAM.

        Returns
        -------
        out : dict
            Nested dictionaries where the top level key is the site index,
            the second level key is the variable name, second level value is
            the output variable value.
        """

        out = {}

        profiles = cls._get_cf_profiles(points_control.sites, cf_file, year)

        for i, site in enumerate(points_control.sites):
            # get SAM inputs from project_points based on the current site
            _, inputs = points_control.project_points[site]

            # ensure that site-specific data is not persisted to other sites
            site_inputs = deepcopy(inputs)

            # set the generation profile as an input.
            site_inputs = cls._make_gen_profile(i, site, profiles, site_df,
                                                site_inputs)

            out[site] = super().reV_run(site, site_df, site_inputs,
                                        output_request)

        return out



# -*- coding: utf-8 -*-
# pylint: disable=no-member
"""
reV-NRWAL analysis module.

This module runs reV data through the NRWAL compute library. This code was
first developed to use a custom offshore wind LCOE equation library but has
since been refactored to analyze any equation library in NRWAL.

Everything in this module operates on the spatiotemporal resolution of the reV
generation output file. This is usually the wind or solar resource resolution
but could be the supply curve resolution after representative profiles is run.
"""
import logging
from warnings import warn

import numpy as np
import pandas as pd

from reV.generation.generation import Gen
from reV.handlers.outputs import Outputs
from reV.utilities import SiteDataField, ResourceMetaField, log_versions
from reV.utilities.exceptions import (
    DataShapeError,
    OffshoreWindInputError,
    OffshoreWindInputWarning,
)

logger = logging.getLogger(__name__)


class RevNrwal:
    """RevNrwal"""

    DEFAULT_META_COLS = (SiteDataField.CONFIG, )
    """Columns from the `site_data` table to join to the output meta data"""

    def __init__(self, gen_fpath, site_data, sam_files, nrwal_configs,
                 output_request, save_raw=True,
                 meta_gid_col=str(ResourceMetaField.GID),  # str() to fix docs
                 site_meta_cols=None):
        """Framework to handle reV-NRWAL analysis.

        ``reV`` NRWAL analysis runs ``reV`` data through the NRWAL
        compute library. Everything in this module operates on the
        spatiotemporal resolution of the ``reV`` generation output file
        (usually the wind or solar resource resolution but could also be
        the supply curve resolution after representative profiles is
        run).

        Parameters
        ----------
        gen_fpath : str
            Full filepath to HDF5 file with ``reV`` generation or
            rep_profiles output. Anything in the `output_request` input
            is added to and/or manipulated within this file.

            .. Note:: If executing ``reV`` from the command line, this
              input can also be ``"PIPELINE"`` to parse the output of
              one of the previous step and use it as input to this call.
              However, note that duplicate executions of ``reV``
              commands prior to this one within the pipeline may
              invalidate this parsing, meaning the `gen_fpath` input
              will have to be specified manually.

        site_data : str | pd.DataFrame
            Site-specific input data for NRWAL calculation.If this input
            is a string, it should be a path that points to a CSV file.
            Otherwise, this input should be a DataFrame with
            pre-extracted site data. Rows in this table should match
            the `meta_gid_col` in the `gen_fpath` meta data input
            sites via a ``gid`` column. A ``config`` column must also be
            provided that corresponds to the `nrwal_configs` input. Only
            sites with a gid in this file's ``gid`` column will be run
            through NRWAL.
        sam_files : dict | str
            A dictionary mapping SAM input configuration ID(s) to SAM
            configuration(s). Keys are the SAM config ID(s) which
            correspond to the keys in the `nrwal_configs` input. Values
            for each key are either a path to a corresponding SAM
            config file or a full dictionary of SAM config inputs. For
            example::

                sam_files = {
                    "default": "/path/to/default/sam.json",
                    "onshore": "/path/to/onshore/sam_config.yaml",
                    "offshore": {
                        "sam_key_1": "sam_value_1",
                        "sam_key_2": "sam_value_2",
                        ...
                    },
                    ...
                }

            This input can also be a string pointing to a single SAM
            config file. In this case, the ``config`` column of the
            CSV points input should be set to ``None`` or left out
            completely. See the documentation for the ``reV`` SAM class
            (e.g. :class:`reV.SAM.generation.WindPower`,
            :class:`reV.SAM.generation.PvWattsv8`,
            :class:`reV.SAM.generation.Geothermal`, etc.) for
            documentation on the allowed and/or required SAM config file
            inputs.
        nrwal_configs : dict
            A dictionary mapping SAM input configuration ID(s) to NRWAL
            configuration(s). Keys are the SAM config ID(s) which
            correspond to the keys in the `sam_files` input. Values
            for each key are either a path to a corresponding NRWAL YAML
            or JSON config file or a full dictionary of NRWAL config
            inputs. For example::

                nrwal_configs = {
                    "default": "/path/to/default/nrwal.json",
                    "onshore": "/path/to/onshore/nrwal_config.yaml",
                    "offshore": {
                        "nrwal_key_1": "nrwal_value_1",
                        "nrwal_key_2": "nrwal_value_2",
                        ...
                    },
                    ...
                }

        output_request : list | tuple
            List of output dataset names to be written to the
            `gen_fpath` file. Any key from the NRWAL configs or any of
            the inputs (site_data or sam_files) is available to be
            exported as an output dataset. If you want to manipulate a
            dset like ``cf_mean`` from `gen_fpath` and include it in the
            `output_request`, you should set ``save_raw=True`` and then
            use ``cf_mean_raw`` in the NRWAL equations as the input.
            This allows you to define an equation in the NRWAL configs
            for a manipulated ``cf_mean`` output that can be included in
            the `output_request` list.
        save_raw : bool, optional
            Flag to save an initial ("raw") copy of input datasets from
            `gen_fpath` that are also part of the `output_request`. For
            example, if you request ``cf_mean`` in output_request but
            also manipulate the ``cf_mean`` dataset in the NRWAL
            equations, the original ``cf_mean`` will be archived under
            the ``cf_mean_raw`` dataset in `gen_fpath`.
            By default, ``True``.
        meta_gid_col : str, optional
            Column label in the source meta data from `gen_fpath` that
            contains the unique gid identifier. This will be joined to
            the site_data ``gid`` column. By default, ``"gid"``.
        site_meta_cols : list | tuple, optional
            Column labels from `site_data` to be added to the meta data
            table in `gen_fpath`. If ``None``, only the columns in
            :attr:`DEFAULT_META_COLS` will be added. Any columns
            requested via this input will be considered *in addition to*
            the :attr:`DEFAULT_META_COLS`. By default, ``None``.
        """

        log_versions(logger)

        # delayed NRWAL import to cause less errors with old reV installs
        # if not running nrwal.
        from NRWAL import NrwalConfig

        self._meta_gid_col = meta_gid_col
        self._gen_fpath = gen_fpath
        self._site_data = site_data
        self._output_request = output_request
        self._meta_out = None
        self._time_index = None
        self._save_raw = save_raw
        self._nrwal_inputs = self._out = None

        self._nrwal_configs = {
            k: NrwalConfig(v) for k, v in nrwal_configs.items()
        }

        self._site_meta_cols = site_meta_cols
        if self._site_meta_cols is None:
            self._site_meta_cols = list(self.DEFAULT_META_COLS)
        else:
            self._site_meta_cols = list(self._site_meta_cols)
            self._site_meta_cols += list(self.DEFAULT_META_COLS)
            self._site_meta_cols = list(set(self._site_meta_cols))

        self._site_data = self._parse_site_data()
        self._meta_source = self._parse_gen_data()
        self._analysis_gids, self._site_data = self._parse_analysis_gids()

        pc = Gen.get_pc(
            self._site_data[[SiteDataField.GID, SiteDataField.CONFIG]],
            points_range=None, sam_configs=sam_files, tech='windpower')
        self._project_points = pc.project_points

        self._sam_sys_inputs = self._parse_sam_sys_inputs()
        meta_gids = self.meta_source[self._meta_gid_col].values
        logger.info(
            'Finished initializing NRWAL analysis module for "{}" '
            "{} through {} with {} total generation points and "
            "{} NRWAL analysis points.".format(
                self._meta_gid_col,
                meta_gids.min(),
                meta_gids.max(),
                len(self.meta_source),
                len(self.analysis_gids),
            )
        )

    def _parse_site_data(self, required_columns=(SiteDataField.GID,
                                                 SiteDataField.CONFIG)):
        """Parse the site-specific spatial input data file

        Parameters
        ----------
        required_columns : tuple | list
            List of column names that must be in the site_data in
            order to run the reV NRWAL module.

        Returns
        -------
        site_data : pd.DataFrame
            Dataframe of extracted site_data. Each row is an analysis point and
            columns are spatial data inputs.
        """

        if isinstance(self._site_data, str):
            self._site_data = pd.read_csv(self._site_data)

        if "dist_l_to_ts" in self._site_data:
            if self._site_data["dist_l_to_ts"].sum() > 0:
                w = (
                    'Possible incorrect Offshore data input! "dist_l_to_ts" '
                    "(distance land to transmission) input is non-zero. "
                    "Most reV runs set this to zero and input the cost "
                    "of transmission from landfall tie-in to "
                    "transmission feature in the supply curve module."
                )
                logger.warning(w)
                warn(w, OffshoreWindInputWarning)

        for c in required_columns:
            if c not in self._site_data:
                msg = 'Did not find required "{}" column in site_data!'.format(
                    c
                )
                logger.error(msg)
                raise KeyError(msg)

        self._site_data = self._site_data.sort_values(SiteDataField.GID)

        return self._site_data

    def _parse_gen_data(self):
        """Parse generation data and get meta data

        Returns
        -------
        meta : pd.DataFrame
            Full meta data from gen_fpath.
        """

        with Outputs(self._gen_fpath, mode="r") as out:
            meta = out.meta

        msg = (
            'Could not find "{}" column in source generation h5 file '
            "meta data! Available cols: {}".format(
                self._meta_gid_col, meta.columns.values.tolist()
            )
        )
        assert self._meta_gid_col in meta, msg

        # currently an assumption of sorted gids in the reV gen output
        msg = "Source capacity factor meta data is not ordered!"
        meta_gids = list(meta[self._meta_gid_col])
        assert meta_gids == sorted(meta_gids), msg

        return meta

    def _parse_analysis_gids(self):
        """Check the intersection of the generation gids and the site_data
        input gids.

        Returns
        -------
        analysis_gids : np.ndarray
            Array indicating which sites in the source meta data to process
            with NRWAL. This is the intersection of the gids in the generation
            meta data and the gids in the site_data input.
        site_data : pd.DataFrame
            The site_data table reduced to only those gids that are in the
            analysis_gids
        """

        meta_gids = self.meta_source[self._meta_gid_col].values

        missing = ~np.isin(meta_gids, self._site_data[SiteDataField.GID])
        if any(missing):
            msg = (
                "{} sites from the generation meta data input were "
                'missing from the "site_data" input and will not be '
                "run through NRWAL: {}".format(
                    missing.sum(), meta_gids[missing]
                )
            )
            logger.info(msg)

        missing = ~np.isin(self._site_data[SiteDataField.GID], meta_gids)
        if any(missing):
            missing = self._site_data[SiteDataField.GID].values[missing]
            msg = ('{} sites from the "site_data" input were missing from the '
                   'generation meta data and will not be run through NRWAL: {}'
                   .format(len(missing), missing))
            logger.info(msg)

        analysis_gids = (set(meta_gids)
                         & set(self._site_data[SiteDataField.GID]))
        analysis_gids = np.array(sorted(list(analysis_gids)))

        # reduce the site data table to only those sites being analyzed
        mask = np.isin(self._site_data[SiteDataField.GID], meta_gids)
        self._site_data = self._site_data[mask]

        return analysis_gids, self._site_data

    def _parse_sam_sys_inputs(self):
        """Get the SAM system inputs dict from project points.

        Returns
        -------
        system_inputs : pd.DataFrame
            DataFrame of SAM config inputs (columns) for every active nrwal
            analysis gid (row). Index is resource gids and there is also a
            column "gid" with the copied gids.
        """

        system_inputs = {}

        for gid in self.analysis_gids:
            system_inputs[gid] = self._project_points[gid][1]

        system_inputs = pd.DataFrame(system_inputs).T
        system_inputs = system_inputs.sort_index()
        system_inputs[SiteDataField.GID] = system_inputs.index.values
        system_inputs.index.name = SiteDataField.GID
        mask = system_inputs[SiteDataField.GID].isin(self.analysis_gids)
        system_inputs = system_inputs[mask]

        return system_inputs

    def _init_outputs(self):
        """Initialize a dictionary of outputs with dataset names as keys and
        numpy arrays as values. All datasets are initialized as 1D arrays and
        must be overwritten if found to be 2D. Only active analysis sites will
        have data in the output, sites that were not found in the site_data
        "gid" column will not have data in these output arrays

        Returns
        -------
        out : dict
            Dictionary of output data
        """
        out = {}

        for key in self._output_request:
            out[key] = np.full(
                len(self.analysis_gids), np.nan, dtype=np.float32
            )

            if key in self.gen_dsets and not self._save_raw:
                msg = (
                    'Output request "{0}" was also found in '
                    "the source gen file but save_raw=False! If "
                    "you are manipulating this "
                    "dset, make sure you set save_raw=False "
                    'and reference "{0}_raw" as the '
                    'input in the NRWAL equations and then define "{0}" '
                    "as the final manipulated dataset.".format(key)
                )
                logger.warning(msg)
                warn(msg)
            elif key in self.gen_dsets:
                msg = (
                    'Output request "{0}" was also found in '
                    "the source gen file. If you are manipulating this "
                    'dset, make sure you reference "{0}_raw" as the '
                    'input in the NRWAL equations and then define "{0}" '
                    "as the final manipulated dataset.".format(key)
                )
                logger.info(msg)

            if key in self._nrwal_inputs:
                out[key] = self._nrwal_inputs[key]

        return out

    def _preflight_checks(self):
        """Run some preflight checks on the offshore inputs"""
        sam_files = {
            k: v
            for k, v in self._project_points.sam_inputs.items()
            if k in self._nrwal_configs
        }

        for cid, sys_in in sam_files.items():
            loss1 = sys_in.get("wind_farm_losses_percent", 0)
            loss2 = sys_in.get("turb_generic_loss", 0)
            if loss1 != 0 or loss2 != 0:
                msg = (
                    'Wind farm loss for config "{}" is not 0. When using '
                    "NRWAL for offshore analysis, consider using gross "
                    "capacity factors from reV generation and applying "
                    "spatially dependent losses from the NRWAL equations"
                    .format(cid)
                )
                logger.info(msg)

        available_ids = list(self._nrwal_configs.keys())
        requested_ids = list(self._site_data[SiteDataField.CONFIG].values)
        missing = set(requested_ids) - set(available_ids)
        if any(missing):
            msg = (
                "The following config ids were requested in the offshore "
                "data input but were not available in the NRWAL config "
                "input dict: {}".format(missing)
            )
            logger.error(msg)
            raise OffshoreWindInputError(msg)

        check_gid_order = (self._site_data[SiteDataField.GID].values
                           == self._sam_sys_inputs[SiteDataField.GID].values)
        msg = 'NRWAL site_data and system input dataframe had bad order'
        assert (check_gid_order).all(), msg

        missing = [c for c in self._site_meta_cols if c not in self._site_data]
        if any(missing):
            msg = (
                "Could not find requested NRWAL site data pass through "
                "columns in offshore input data: {}".format(missing)
            )
            logger.error(msg)
            raise OffshoreWindInputError(msg)

    def _get_input_data(self):
        """Get all the input data from the site_data, SAM system configs, and
        generation h5 file, formatted together in one dictionary for NRWAL.

        Returns
        -------
        nrwal_inputs : dict
            Dictionary mapping required NRWAL input variable names (keys) to 1
            or 2D arrays of inputs for all the analysis_gids
        """

        logger.info("Setting up input data for NRWAL...")

        # preconditions for this to work properly
        assert len(self._site_data) == len(self.analysis_gids)
        assert len(self._sam_sys_inputs) == len(self.analysis_gids)

        all_required = []
        for config_id, nrwal_config in self._nrwal_configs.items():
            all_required += list(nrwal_config.required_inputs)
            all_required = list(set(all_required))

            missing_vars = [
                var
                for var in nrwal_config.required_inputs
                if var not in self._site_data
                and var not in self.meta_source
                and var not in self._sam_sys_inputs
                and var not in self.gen_dsets
            ]

            if any(missing_vars):
                msg = (
                    "Could not find required input variables {} "
                    'for NRWAL config "{}" in either the offshore '
                    "data or the SAM system data!".format(
                        missing_vars, config_id
                    )
                )
                logger.error(msg)
                raise OffshoreWindInputError(msg)

        meta_data_vars = [
            var for var in all_required if var in self.meta_source
        ]
        logger.info(
            "Pulling the following inputs from the gen meta data: {}".format(
                meta_data_vars
            )
        )
        nrwal_inputs = {
            var: self.meta_source[var].values[self.analysis_mask]
            for var in meta_data_vars
        }

        site_data_vars = [var for var in all_required
                          if var in self._site_data
                          and var not in nrwal_inputs]
        site_data_vars.append(SiteDataField.CONFIG)
        logger.info('Pulling the following inputs from the site_data input: {}'
                    .format(site_data_vars))
        for var in site_data_vars:
            nrwal_inputs[var] = self._site_data[var].values

        sam_sys_vars = [
            var
            for var in all_required
            if var in self._sam_sys_inputs and var not in nrwal_inputs
        ]
        logger.info(
            "Pulling the following inputs from the SAM system "
            "configs: {}".format(sam_sys_vars)
        )
        for var in sam_sys_vars:
            nrwal_inputs[var] = self._sam_sys_inputs[var].values

        gen_vars = [
            var
            for var in all_required
            if var in self.gen_dsets and var not in nrwal_inputs
        ]
        logger.info(
            "Pulling the following inputs from the generation "
            "h5 file: {}".format(gen_vars)
        )
        with Outputs(self._gen_fpath, mode="r") as f:
            source_gids = self.meta_source[self._meta_gid_col]
            gen_gids = np.where(source_gids.isin(self.analysis_gids))[0]
            for var in gen_vars:
                shape = f.shapes[var]
                if len(shape) == 1:
                    nrwal_inputs[var] = f[var, gen_gids]
                elif len(shape) == 2:
                    nrwal_inputs[var] = f[var, :, gen_gids]
                else:
                    msg = (
                        'Data shape for "{}" must be 1 or 2D but '
                        "received: {}".format(var, shape)
                    )
                    logger.error(msg)
                    raise DataShapeError(msg)

        logger.info("Finished setting up input data for NRWAL!")

        return nrwal_inputs

    @property
    def time_index(self):
        """Get the source time index."""
        if self._time_index is None:
            with Outputs(self._gen_fpath, mode="r") as out:
                self._time_index = out.time_index

        return self._time_index

    @property
    def gen_dsets(self):
        """Get the available datasets from the gen source file"""
        with Outputs(self._gen_fpath, mode="r") as out:
            dsets = out.dsets

        return dsets

    @property
    def meta_source(self):
        """Get the full meta data (onshore + offshore)"""
        return self._meta_source

    @property
    def meta_out(self):
        """Get the combined onshore and offshore meta data."""
        if self._meta_out is None:
            self._meta_out = self._meta_source.copy()
            for col in self._site_meta_cols:
                data = self._nrwal_inputs[col]
                self._meta_out.loc[self.analysis_mask, col] = data

        return self._meta_out

    @property
    def analysis_mask(self):
        """Get a boolean array to mask the source generation meta data where
        True is sites that are to be analyzed by NRWAL.

        Returns
        -------
        np.ndarray
        """
        mask = np.isin(
            self.meta_source[self._meta_gid_col], self.analysis_gids
        )
        return mask

    @property
    def analysis_gids(self):
        """Get an array of gids from the source generation meta data that are
        to-be analyzed by nrwal.

        Returns
        -------
        np.ndarray
        """
        return self._analysis_gids

    @property
    def outputs(self):
        """Get a dict of NRWAL outputs. Only active analysis sites will have
        data in the output, sites that were not found in the site_data "gid"
        column will not have data in these output arrays"""
        return self._out

    def _save_nrwal_out(self, name, nrwal_out, output_mask):
        """Save a dataset from the nrwal_out dictionary to the self._out
        attribute

        Parameters
        ----------
        name : str
            Dataset name of the nrwal output to be saved.
        nrwal_out : dict
            Output dictionary from a successfully evaluated NrwalConfig object
            containing the dataset with the input name
        output_mask : np.ndarray
            Boolean array showing which gids in self.analysis_gids should be
            assigned data from this NRWAL output. If not all true, there are
            probably multiple NrwalConfig objects that map to different sets of
            gids.
        """
        value = nrwal_out[name]
        value = self._value_to_array(value, name)

        if len(value.shape) == 1:
            self._out[name][output_mask] = value[output_mask]

        elif len(value.shape) == 2:
            if len(self._out[name].shape) == 1:
                if not all(np.isnan(self._out[name])):
                    msg = (
                        'Output dataset "{}" was initialized as 1D but was '
                        "later found to be 2D but was not all NaN!".format(
                            name
                        )
                    )
                    logger.error(msg)
                    raise DataShapeError(msg)

                # re-initialize the dataset as 2D now that we
                # know what the output looks like
                out_shape = (len(self.time_index), len(self.analysis_gids))
                self._out[name] = np.full(out_shape, np.nan, dtype=np.float32)

            self._out[name][:, output_mask] = value[:, output_mask]

        else:
            msg = (
                'Could not make sense of NRWAL output "{}" '
                "with shape {}".format(name, value.shape)
            )
            logger.error(msg)
            raise DataShapeError(msg)

    def _save_nrwal_misc(self, name, nrwal_config, output_mask):
        """Save miscellaneous output requests from a NRWAL config object (not
        NRWAL output dictionary) to the self._out attribute.

        Parameters
        ----------
        name : str
            Dataset name of the nrwal output to be saved.
        nrwal_config : NrwalConfig
            NrwalConfig object containing NRWAL Equation objects or something
            else that is to be exported to the outputs
        output_mask : np.ndarray
            Boolean array showing which gids in self.analysis_gids should be
            assigned data from this NRWAL output. If not all true, there are
            probably multiple NrwalConfig objects that map to different sets of
            gids.
        """

        from NRWAL import Equation

        value = nrwal_config[name]

        if isinstance(value, Equation):
            msg = (
                'Cannot retrieve Equation "{}" from NRWAL. '
                "Must be a number!".format(name)
            )
            assert not any(value.variables), msg
            value = value.eval()

        value = self._value_to_array(value, name)
        self._out[name][output_mask] = value[output_mask]

    def _value_to_array(self, value, name):
        """Turn the input into numpy array if it isn't already."""
        if np.issubdtype(type(value), np.number):
            value *= np.ones(len(self.analysis_gids), dtype=np.float32)

        if not isinstance(value, np.ndarray):
            msg = (
                'NRWAL key "{}" returned bad type of "{}", needs to be '
                "numeric or an output array.".format(name, type(value))
            )
            logger.error(msg)
            raise TypeError(msg)
        return value

    def run_nrwal(self):
        """Run analysis via the NRWAL analysis library"""

        self._preflight_checks()
        self.save_raw_dsets()
        self._nrwal_inputs = self._get_input_data()
        self._out = self._init_outputs()

        for i, (cid, nrwal_config) in enumerate(self._nrwal_configs.items()):
            output_mask = self._site_data[SiteDataField.CONFIG].values == cid
            logger.info('Running NRWAL config {} of {}: "{}" and applying '
                        'to {} out of {} total sites'
                        .format(i + 1, len(self._nrwal_configs), cid,
                                output_mask.sum(), len(output_mask)))

            nrwal_out = nrwal_config.eval(inputs=self._nrwal_inputs)

            # pylint: disable=C0201
            for name in self._out.keys():
                if name in nrwal_out:
                    self._save_nrwal_out(name, nrwal_out, output_mask)

                elif name in nrwal_config.keys():
                    self._save_nrwal_misc(name, nrwal_config, output_mask)

                elif name not in self._nrwal_inputs:
                    msg = (
                        'Could not find "{}" in the output dict of NRWAL '
                        "config {}".format(name, cid)
                    )
                    logger.warning(msg)
                    warn(msg)

    def check_outputs(self):
        """Check the nrwal outputs for nan values and raise errors if found."""
        for name, arr in self._out.items():
            if np.isnan(arr).all():
                msg = (
                    'Output array "{}" is all NaN! Probably was not '
                    "found in the available NRWAL keys.".format(name)
                )
                logger.warning(msg)
                warn(msg)
            elif np.isnan(arr).any():
                mask = np.isnan(arr)
                nan_meta = self.meta_source[self.analysis_mask][mask]
                nan_gids = nan_meta[self._meta_gid_col].values
                msg = (
                    "NaN values ({} out of {}) persist in NRWAL "
                    'output "{}"!'.format(np.isnan(arr).sum(), len(arr), name)
                )
                logger.warning(msg)
                logger.warning(
                    "This is the NRWAL meta that is causing NaN "
                    "outputs: {}".format(nan_meta)
                )
                logger.warning(
                    "These are the resource gids causing NaN "
                    "outputs: {}".format(nan_gids)
                )
                warn(msg)

    def save_raw_dsets(self):
        """If requested by save_raw=True, archive raw datasets that exist in
        the gen_fpath file and are also requested in the output_request"""
        if self._save_raw:
            with Outputs(self._gen_fpath, "a") as f:
                for dset in self._output_request:
                    dset_raw = "{}_raw".format(dset)
                    if dset in f and dset_raw not in f:
                        logger.info(
                            'Saving raw data from "{}" to "{}"'.format(
                                dset, dset_raw
                            )
                        )
                        f._add_dset(
                            dset_raw,
                            f[dset],
                            f.dtypes[dset],
                            attrs=f.attrs[dset],
                        )

    def write_to_gen_fpath(self):
        """Save NRWAL outputs to input generation fpath file.

        Returns
        -------
        str
            Path to output file.
        """

        logger.info("Writing NRWAL outputs to: {}".format(self._gen_fpath))
        write_all = self.analysis_mask.all()

        with Outputs(self._gen_fpath, "a") as f:
            meta_attrs = f.attrs["meta"]
            del f._h5["meta"]
            f._set_meta("meta", self.meta_out, attrs=meta_attrs)

            for dset, arr in self._out.items():
                if len(arr.shape) == 1:
                    data = np.full(
                        len(self.meta_source), np.nan, dtype=np.float32
                    )
                else:
                    full_shape = (len(self.time_index), len(self.meta_source))
                    data = np.full(full_shape, np.nan, dtype=np.float32)

                dset_attrs = {"scale_factor": 1}
                dset_dtype = np.float32
                if dset in f.dsets:
                    logger.info(
                        'Found "{}" in file, loading data and '
                        "overwriting data for {} out of {} sites.".format(
                            dset,
                            self.analysis_mask.sum(),
                            len(self.analysis_mask),
                        )
                    )
                    dset_attrs = f.attrs[dset]
                    dset_dtype = f.dtypes[dset]
                    if not write_all:
                        data = f[dset]

                if len(arr.shape) == 1:
                    data[self.analysis_mask] = arr
                else:
                    data[:, self.analysis_mask] = arr

                logger.info(
                    'Writing final "{}" to: {}'.format(dset, self._gen_fpath)
                )
                f._add_dset(dset, data, dset_dtype, attrs=dset_attrs)

        logger.info(
            "Finished writing NRWAL outputs to: {}".format(self._gen_fpath)
        )
        return self._gen_fpath

    def write_meta_to_csv(self, out_fpath=None):
        """Combine NRWAL outputs with meta and write to output csv.

        Parameters
        ----------
        out_fpath : str, optional
            Full path to output NRWAL CSV file. The file path does not
            need to include file ending - it will be added automatically
            if missing. If ``None``, the generation HDF5 filepath will
            be converted to a CSV out path by replacing the ".h5" file
            ending with ".csv". By default, ``None``.

        Returns
        -------
        str
            Path to output file.
        """
        if out_fpath is None:
            out_fpath = self._gen_fpath.replace(".h5", ".csv")
        elif not out_fpath.endswith(".csv"):
            out_fpath = "{}.csv".format(out_fpath)

        logger.info("Writing NRWAL outputs to: {}".format(out_fpath))
        meta_out = self.meta_out[self.analysis_mask].copy()

        for dset, arr in self._out.items():
            if len(arr.shape) != 1 or arr.shape[0] != meta_out.shape[0]:
                msg = (
                    "Skipping output {!r}: shape {} cannot be combined "
                    "with meta of shape {}!".format(
                        dset, arr.shape, meta_out.shape
                    )
                )
                logger.warning(msg)
                warn(msg)
                continue
            meta_out[dset] = arr

        meta_out.to_csv(out_fpath, index=False)
        logger.info("Finished writing NRWAL outputs to: {}".format(out_fpath))
        return out_fpath

    def run(self, csv_output=False, out_fpath=None):
        """Run NRWAL analysis.

        Parameters
        ----------
        csv_output : bool, optional
            Option to write H5 file meta + all requested outputs to
            CSV file instead of storing in the HDF5 file directly. This
            can be useful if the same HDF5 file is used for multiple
            sets of NRWAL runs. Note that all requested output datasets
            must be 1-dimensional in order to fir within the CSV output.

            .. Important:: This option is not compatible with
              ``save_raw=True``. If you set ``csv_output=True``, then
              the `save_raw` option is forced to be ``False``.
              Therefore, make sure that you do not have any references
              to "input_dataset_name_raw" in your NRWAL config. If you
              need to manipulate an input dataset, save it to a
              different output name in the NRWAL config or manually add
              an "input_dataset_name_raw" dataset to your generation
              HDF5 file before running NRWAL.

            By default, ``False``.
        out_fpath : str, optional
            This option has no effect if ``csv_output=False``.
            Otherwise, this should be the full path to output NRWAL CSV
            file. The file path does not need to include file ending -
            it will be added automatically if missing. If ``None``, the
            generation HDF5 filepath will be converted to a CSV out path
            by replacing the ".h5" file ending with ".csv".
            By default, ``None``.

        Returns
        -------
        str
            Path to output file.
        """
        if csv_output and self._save_raw:
            msg = (
                "`save_raw` option not allowed with `csv_output`. Setting"
                "`save_raw=False`"
            )
            logger.warning(msg)
            warn(msg)
            self._save_raw = False

        if any(self.analysis_gids):
            self.run_nrwal()
            self.check_outputs()
            if csv_output:
                out_fp = self.write_meta_to_csv(out_fpath)
            else:
                out_fp = self.write_to_gen_fpath()

        logger.info("NRWAL module complete!")

        return out_fp
